import copy
import hashlib
import logging
import os
import sys
import traceback
import warnings
from functools import cmp_to_key
from typing import Dict
from typing import List
from typing import Optional
from typing import Union
from urllib.parse import parse_qs
from urllib.parse import unquote
from urllib.parse import urljoin
from urllib.parse import urlparse

from jwkest import jws

from oic import rndstr
from oic.exception import AuthzError
from oic.exception import FailedAuthentication
from oic.exception import InvalidRequest
from oic.exception import MissingParameter
from oic.exception import ParameterError
from oic.exception import RedirectURIError
from oic.exception import UnknownClient
from oic.exception import UnSupported
from oic.exception import URIError
from oic.oauth2 import ErrorResponse
from oic.oauth2 import Server
from oic.oauth2 import error_response
from oic.oauth2 import none_response
from oic.oauth2 import redirect_authz_error
from oic.oauth2.message import AccessTokenResponse
from oic.oauth2.message import AuthorizationResponse
from oic.oauth2.message import Message
from oic.oauth2.message import MissingRequiredAttribute
from oic.oauth2.message import MissingRequiredValue
from oic.oauth2.message import OauthMessageFactory
from oic.oauth2.message import TokenErrorResponse
from oic.oauth2.message import add_non_standard
from oic.oauth2.message import by_schema
from oic.utils import sort_sign_alg
from oic.utils.authn.client import AuthnFailure
from oic.utils.authn.user import NoSuchAuthentication
from oic.utils.authn.user import TamperAllert
from oic.utils.authn.user import ToOld
from oic.utils.clientdb import BaseClientDatabase
from oic.utils.http_util import OAUTH2_NOCACHE_HEADERS
from oic.utils.http_util import BadRequest
from oic.utils.http_util import CookieDealer
from oic.utils.http_util import Response
from oic.utils.http_util import SeeOther
from oic.utils.http_util import Unauthorized
from oic.utils.http_util import make_cookie
from oic.utils.keyio import KeyJar
from oic.utils.sanitize import sanitize
from oic.utils.sdb import AccessCodeUsed
from oic.utils.session_backend import AuthnEvent
from oic.utils.settings import OauthProviderSettings
from oic.utils.settings import PyoidcSettings

__author__ = "rohe0002"

logger = logging.getLogger(__name__)

STR = 5 * "_"

CAPABILITIES = {
    "response_types_supported": ["code", "token"],
    "response_modes_supported": ["query", "fragment", "form_post"],
    "grant_types_supported": [
        "authorization_code",
        "implicit",
        "urn:ietf:params:oauth:grant-type:jwt-bearer",
    ],
}

AUTH_METHODS_SUPPORTED = [
    "client_secret_post",
    "client_secret_basic",
    "client_secret_jwt",
    "private_key_jwt",
]


class Endpoint(object):
    """
    Endpoint class.

    @var etype: Endpoint type
    @url: Relative part of the url (will be joined with server.baseurl)
    """

    etype = ""
    url = ""

    def __init__(self, func=None):
        self.func = func

    def __call__(self, *args, **kwargs):
        return self.func(*args, **kwargs)


class AuthorizationEndpoint(Endpoint):
    etype = "authorization"
    url = "authorization"


class TokenEndpoint(Endpoint):
    etype = "token"
    url = "token"


def endpoint_ava(endp, baseurl):
    key = "{}_endpoint".format(endp.etype)
    val = urljoin(baseurl, endp.url)
    return {key: val}


def code_response(**kwargs):
    aresp = AuthorizationResponse()
    _areq = kwargs["areq"]
    try:
        aresp["state"] = _areq["state"]
    except KeyError:
        pass
    aresp["code"] = kwargs["scode"]
    # TODO Add 'iss' and 'client_id'
    if kwargs["myself"]:
        aresp["iss"] = kwargs["myself"]
    aresp["client_id"] = _areq["client_id"]
    add_non_standard(_areq, aresp)
    return aresp


def token_response(**kwargs):
    _areq = kwargs["areq"]
    _scode = kwargs["scode"]
    _sdb = kwargs["sdb"]
    _dic = _sdb.upgrade_to_token(_scode, issue_refresh=False)

    aresp = AccessTokenResponse(**by_schema(AccessTokenResponse, **_dic))

    try:
        aresp["state"] = _areq["state"]
    except KeyError:
        pass

    add_non_standard(_areq, aresp)
    return aresp


def location_url(response_type, redirect_uri, query):
    if response_type in [["code"], ["token"], ["none"]]:
        return "%s?%s" % (redirect_uri, query)
    else:
        return "%s#%s" % (redirect_uri, query)


def max_age(areq):
    try:
        return areq["request"]["max_age"]
    except KeyError:
        try:
            return areq["max_age"]
        except KeyError:
            return 0


def re_authenticate(areq, authn):
    if "prompt" in areq and "login" in areq["prompt"]:
        if authn.done(areq):
            return True

    return False


DELIM = "]["


class Provider(object):
    endp = [AuthorizationEndpoint, TokenEndpoint]

    def __init__(
        self,
        name,
        sdb,
        cdb,
        authn_broker,
        authz,
        client_authn,
        symkey=None,
        urlmap=None,
        iv=0,
        default_scope="",
        verify_ssl=None,
        default_acr="",
        keyjar=None,
        baseurl="",
        server_cls=Server,
        client_cert=None,
        message_factory=OauthMessageFactory,
        capabilities=None,
        jwks_uri="",
        settings: Optional[PyoidcSettings] = None,
    ):
        self.settings = settings or OauthProviderSettings()
        if verify_ssl is not None:
            warnings.warn(
                "`verify_ssl` is deprecated, please use `settings` instead if you need to set a non-default value.",
                DeprecationWarning,
                stacklevel=2,
            )
            self.settings.verify_ssl = verify_ssl
        if client_cert is not None:
            warnings.warn(
                "`client_cert` is deprecated, please use `settings` instead if you need to set a non-default value.",
                DeprecationWarning,
                stacklevel=2,
            )
            self.settings.client_cert = client_cert

        self.name = name
        self.sdb = sdb
        if not isinstance(cdb, BaseClientDatabase):
            warnings.warn(
                "ClientDatabase should be an instance of "
                "oic.utils.clientdb.BaseClientDatabase to ensure proper API."
            )
        self.cdb = cdb
        self.server = server_cls(
            keyjar=keyjar,
            message_factory=message_factory,
            settings=self.settings,
        )

        self.authn_broker = authn_broker
        if authn_broker is None:
            # default cookie function
            self.cookie_func = CookieDealer(srv=self).create_cookie
        else:
            self.cookie_func = self.authn_broker[0][0].create_cookie
            for item in self.authn_broker:
                item.srv = self

        self.authz = authz
        self.client_authn = client_authn
        self.symkey = symkey
        self.seed = rndstr().encode("utf-8")
        self.iv = iv or os.urandom(16)
        self.cookie_name = "pyoidc"
        self.cookie_domain = ""
        self.cookie_path = ""
        self.default_scope = default_scope
        self.sso_ttl = 0
        self.default_acr = default_acr

        if urlmap is None:
            self.urlmap: Dict[str, List[str]] = {}
        else:
            self.urlmap = urlmap

        self.response_type_map = {
            "code": code_response,
            "token": token_response,
            "none": none_response,
        }

        self.session_cookie_name = "pyoic_session"
        self.sso_cookie_name = "pyoidc_sso"
        self.baseurl = baseurl
        self.keyjar: Optional[KeyJar] = None
        self.trace = None
        self.events = None
        self.scopes = ["offline_access"]

        self.jwks_uri = jwks_uri

        if capabilities:
            self.verify_capabilities(capabilities)
            self.capabilities = message_factory.get_response_type(
                "configuration_endpoint"
            )(**capabilities)
        else:
            self.capabilities = self.provider_features()
        self.capabilities["issuer"] = self.name

    @staticmethod
    def input(query="", post=None):
        # Support GET and POST
        if query:
            return query
        elif post:
            return post
        else:
            raise MissingParameter("No input")

    @property
    def default_capabilities(self):
        """Define default capabilities for implementation."""
        return CAPABILITIES

    def endpoints(self):
        return [endp.url for endp in self.endp]

    def _verify_redirect_uri(self, areq):
        """
        Verify that redirect_uri is valid.

        MUST NOT contain a fragment
        MAY contain query component

        :return: An error response if the redirect URI is faulty otherwise None
        """
        try:
            _redirect_uri = unquote(areq["redirect_uri"])

            part = urlparse(_redirect_uri)
            if part.fragment:
                raise URIError("Contains fragment")

            _query = parse_qs(part.query) if part.query else None
            _base = part._replace(query="").geturl()

            match = False
            for regbase, rquery in self.cdb[str(areq["client_id"])]["redirect_uris"]:
                # The URI MUST exactly match one of the Redirection URI
                if _base != regbase:
                    continue

                if not rquery and not _query:
                    match = True
                    break

                if not rquery or not _query:
                    continue

                # every registered query component must exist in the
                # redirect_uri
                is_match_query = True
                for key, vals in _query.items():
                    if key not in rquery:
                        is_match_query = False
                        break

                    for val in vals:
                        if val not in rquery[key]:
                            is_match_query = False
                            break

                    if not is_match_query:
                        break

                if not is_match_query:
                    continue

                match = True
                break

            if not match:
                raise RedirectURIError("Doesn't match any registered uris")
            # ignore query components that are not registered
            return None
        except Exception:
            logger.error("Faulty redirect_uri: %s" % areq["redirect_uri"])
            try:
                _cinfo = self.cdb[str(areq["client_id"])]
            except KeyError:
                try:
                    cid = areq["client_id"]
                except KeyError:
                    logger.error("No client id found")
                    raise UnknownClient("No client_id provided")
                else:
                    logger.info("Unknown client: %s" % cid)
                    raise UnknownClient(areq["client_id"])
            else:
                logger.info("Registered redirect_uris: %s" % sanitize(_cinfo))
                raise RedirectURIError("Faulty redirect_uri: %s" % areq["redirect_uri"])

    def verify_capabilities(self, capabilities) -> bool:
        """
        Verify that what the admin wants the server to do actually can be done by this implementation.

        :param capabilities: The asked for capabilities as a dictionary
        or a ProviderConfigurationResponse instance. The later can be
        treated as a dictionary.
        """
        _pinfo = self.provider_features()
        not_supported: Dict[str, Union[str, List[str]]] = {}
        for key, val in capabilities.items():
            if isinstance(val, str):
                if val not in _pinfo.get(key, ""):
                    not_supported[key] = val
            elif isinstance(val, bool):
                if not _pinfo.get(key) and val:
                    not_supported[key] = ""
            elif isinstance(val, list):
                unsup = []
                for v in val:
                    if v not in _pinfo.get(key, ""):
                        unsup.append(v)
                if unsup:
                    not_supported[key] = unsup
        if not_supported:
            logger.error(
                "Server does not support the following features: %s", not_supported
            )
            return False
        return True

    def provider_features(self, provider_config=None):
        """
        Present what the server capabilities are.

        :return: ProviderConfigurationResponse instance
        """
        pcr_class = self.server.message_factory.get_response_type(
            "configuration_endpoint"
        )

        _provider_info = pcr_class(**self.default_capabilities)
        _provider_info["scopes_supported"] = self.scopes

        sign_algs = list(jws.SIGNER_ALGS.keys())
        sign_algs.remove("none")
        sign_algs = sorted(sign_algs, key=cmp_to_key(sort_sign_alg))

        _pat1 = "{}_endpoint_auth_signing_alg_values_supported"
        _pat2 = "{}_endpoint_auth_methods_supported"
        for typ in ["token", "revocation", "introspection"]:
            _provider_info[_pat1.format(typ)] = sign_algs
            _provider_info[_pat2.format(typ)] = AUTH_METHODS_SUPPORTED

        if provider_config:
            _provider_info.update(provider_config)

        return _provider_info

    def create_providerinfo(self, setup=None):
        """
        Dynamically create the provider info response.

        :param setup:
        :return:
        """
        pcr_class = self.server.message_factory.get_response_type(
            "configuration_endpoint"
        )
        _provider_info = copy.deepcopy(self.capabilities.to_dict())

        if self.jwks_uri and self.keyjar:
            _provider_info["jwks_uri"] = self.jwks_uri

        for endp in self.endp:
            if not self.baseurl.endswith("/"):
                baseurl = self.baseurl + "/"
            else:
                baseurl = self.baseurl
            _provider_info["{}_endpoint".format(endp.etype)] = urljoin(
                baseurl, endp.url
            )

        if setup and isinstance(setup, dict):
            for key in pcr_class.c_param.keys():
                if key in setup:
                    _provider_info[key] = setup[key]

        _provider_info["issuer"] = self.name
        _provider_info["version"] = "3.0"

        return pcr_class(**_provider_info)

    def providerinfo_endpoint(self, handle="", **kwargs):
        _log_info = logger.info

        _log_info("@providerinfo_endpoint")
        try:
            _response = self.create_providerinfo()
            msg = "provider_info_response: {}"
            _log_info(msg.format(sanitize(_response.to_dict())))
            if self.events:
                self.events.store("Protocol response", _response)

            headers = [("Cache-Control", "no-store")]
            if handle:
                (key, timestamp) = handle
                if key.startswith(STR) and key.endswith(STR):
                    cookie = self.cookie_func(
                        key, self.cookie_name, "pinfo", self.sso_ttl
                    )
                    headers.append(cookie)

            resp = Response(
                _response.to_json(), content="application/json", headers=headers
            )
        except Exception:
            message = traceback.format_exception(*sys.exc_info())
            logger.error(message)
            resp = error_response("service_error", message)

        return resp

    def get_redirect_uri(self, areq):
        """
        Verify that the redirect URI is reasonable.

        :param areq: The Authorization request
        :return: Tuple of (redirect_uri, Response instance)
            Response instance is not None of matching redirect_uri failed
        """
        if "redirect_uri" in areq:
            self._verify_redirect_uri(areq)
            uri = areq["redirect_uri"]
        else:
            raise ParameterError(
                "Missing redirect_uri and more than one or none registered"
            )

        return uri

    def pick_auth(self, areq, comparision_type=""):
        """
        Select an authentication method suitable for request.

        :param areq: AuthorizationRequest instance
        :param comparision_type: How to pick the authentication method
        :return: An authentication method and its authn class ref
        """
        if comparision_type == "any":
            return self.authn_broker[0]

        try:
            if len(self.authn_broker) == 1:
                return self.authn_broker[0]
            elif "acr_values" in areq:
                if not comparision_type:
                    comparision_type = "exact"

                if not isinstance(areq["acr_values"], list):
                    areq["acr_values"] = [areq["acr_values"]]

                for acr in areq["acr_values"]:
                    res = self.authn_broker.pick(acr, comparision_type)
                    logger.debug(
                        "Picked AuthN broker for ACR %s: %s" % (str(acr), str(res))
                    )
                    if res:
                        # Return the best guess by pick.
                        return res[0]
            else:  # same as any
                try:
                    acrs = areq["claims"]["id_token"]["acr"]["values"]
                except KeyError:
                    return self.authn_broker[0]
                else:
                    for acr in acrs:
                        res = self.authn_broker.pick(acr, comparision_type)
                        logger.debug(
                            "Picked AuthN broker for ACR %s: %s" % (str(acr), str(res))
                        )
                        if res:
                            # Return the best guess by pick.
                            return res[0]

        except KeyError as exc:
            logger.debug(
                "An error occured while picking the authN broker: %s" % str(exc)
            )

        # return the best I have
        return None, None

    def filter_request(self, req):
        return req

    def auth_init(self, request):
        """
        Start the authentication process.

        :param request: The AuthorizationRequest
        :return:
        """
        request_class = self.server.message_factory.get_request_type(
            "authorization_endpoint"
        )
        logger.debug("Request: '%s'" % sanitize(request))
        # Same serialization used for GET and POST

        try:
            areq = self.server.parse_authorization_request(query=request)
        except (MissingRequiredValue, MissingRequiredAttribute, AuthzError) as err:
            logger.debug("%s" % err)
            areq = request_class()
            areq.lax = True
            if isinstance(request, dict):
                areq.from_dict(request)
            else:
                areq.deserialize(request, "urlencoded")
            try:
                redirect_uri = self.get_redirect_uri(areq)
            except (RedirectURIError, ParameterError, UnknownClient) as err:
                return error_response("invalid_request", "%s" % err)
            try:
                _rtype = areq["response_type"]
            except KeyError:
                _rtype = ["code"]
            try:
                _state = areq["state"]
            except KeyError:
                _state = ""

            return redirect_authz_error(
                "invalid_request", redirect_uri, "%s" % err, _state, _rtype
            )
        except KeyError:
            areq = request_class().deserialize(request, "urlencoded")
            # verify the redirect_uri
            try:
                self.get_redirect_uri(areq)
            except (RedirectURIError, ParameterError) as err:
                return error_response("invalid_request", "%s" % err)
        except Exception as err:
            message = traceback.format_exception(*sys.exc_info())
            logger.error(message)
            logger.debug("Bad request: %s (%s)" % (err, err.__class__.__name__))
            error = ErrorResponse(error="invalid_request", error_description=str(err))
            return BadRequest(error.to_json(), content="application/json")

        if not areq:
            logger.debug("No AuthzRequest")
            return error_response("invalid_request", "Can not parse AuthzRequest")

        if isinstance(areq, Response):
            return areq

        areq = self.filter_request(areq)

        if self.events:
            self.events.store("Protocol request", areq)

        try:
            _cinfo = self.cdb[areq["client_id"]]
        except KeyError:
            logger.error(
                "Client ID ({}) not in client database".format(areq["client_id"])
            )
            return error_response("unauthorized_client", "unknown client")
        else:
            try:
                _registered = [set(rt.split(" ")) for rt in _cinfo["response_types"]]
            except KeyError:
                # If no response_type is registered by the client then we'll
                # code which it the default according to the OIDC spec.
                _registered = [{"code"}]

            _wanted = set(areq["response_type"])
            if _wanted not in _registered:
                return error_response(
                    "invalid_request", "Trying to use unregistered response_typ"
                )

        logger.debug("AuthzRequest: %s" % (sanitize(areq.to_dict()),))
        try:
            redirect_uri = self.get_redirect_uri(areq)
        except (RedirectURIError, ParameterError, UnknownClient) as err:
            return error_response(
                "invalid_request", "{}:{}".format(err.__class__.__name__, err)
            )

        try:
            keyjar = self.keyjar
        except AttributeError:
            keyjar = None

        try:
            # verify that the request message is correct
            areq.verify(keyjar=keyjar, opponent_id=areq["client_id"])
        except (MissingRequiredAttribute, ValueError, MissingRequiredValue) as err:
            return redirect_authz_error("invalid_request", redirect_uri, "%s" % err)

        return {"areq": areq, "redirect_uri": redirect_uri}

    @staticmethod
    def _acr_claims(areq):
        try:
            acrdef = areq["claims"]["id_token"]["acr"]
        except KeyError:
            return None
        else:
            if isinstance(acrdef, dict):
                try:
                    return [acrdef["value"]]
                except KeyError:
                    try:
                        return acrdef["values"]
                    except KeyError:
                        pass

        return None

    def do_auth(self, areq, redirect_uri, cinfo, request, cookie, **kwargs):
        """
        Perform the authentication.

        :param areq:
        :param redirect_uri:
        :param cinfo:
        :param request:
        :param cookie:
        :param authn:
        :param kwargs:
        :return:
        """
        acrs = self._acr_claims(areq)
        if acrs:
            # If acr claims are present the picked acr value MUST match
            # one of the given
            tup = (None, None)
            for acr in acrs:
                res = self.authn_broker.pick(acr, "exact")
                logger.debug(
                    "Picked AuthN broker for ACR %s: %s" % (str(acr), str(res))
                )
                if res:  # Return the best guess by pick.
                    tup = res[0]
                    break
            authn, authn_class_ref = tup
        else:
            authn, authn_class_ref = self.pick_auth(areq)
            if not authn:
                authn, authn_class_ref = self.pick_auth(areq, "better")
                if not authn:
                    authn, authn_class_ref = self.pick_auth(areq, "any")

        if authn is None:
            return redirect_authz_error(
                "access_denied", redirect_uri, return_type=areq["response_type"]
            )

        try:
            try:
                _auth_info = kwargs["authn"]
            except KeyError:
                _auth_info = ""

            if "upm_answer" in areq and areq["upm_answer"] == "true":
                _max_age = 0
            else:
                _max_age = max_age(areq)

            identity, _ts = authn.authenticated_as(
                cookie, authorization=_auth_info, max_age=_max_age
            )
        except (NoSuchAuthentication, TamperAllert):
            identity = None
            _ts = 0
        except ToOld:
            logger.info("Too old authentication")
            identity = None
            _ts = 0
        else:
            logger.info("No active authentication")

        # gather information to be used by the authentication method
        authn_args = {"authn_class_ref": authn_class_ref}
        # Can't be something like JSON because it can't contain '"'
        if isinstance(request, Message):
            authn_args["query"] = request.to_urlencoded()
        elif isinstance(request, dict):
            authn_args["query"] = Message(**request).to_urlencoded()
        else:
            authn_args["query"] = request

        if "req_user" in kwargs:
            authn_args["as_user"] = (kwargs["req_user"],)

        for attr in ["policy_uri", "logo_uri", "tos_uri"]:
            try:
                authn_args[attr] = cinfo[attr]
            except KeyError:
                pass

        for attr in ["ui_locales", "acr_values"]:
            try:
                authn_args[attr] = areq[attr]
            except KeyError:
                pass

        # To authenticate or Not
        if identity is None:  # No!
            if "prompt" in areq and "none" in areq["prompt"]:
                # Need to authenticate but not allowed
                return redirect_authz_error(
                    "login_required", redirect_uri, return_type=areq["response_type"]
                )
            else:
                return authn(**authn_args)
        else:
            if re_authenticate(areq, authn):
                # demand re-authentication
                return authn(**authn_args)
            else:
                # I get back a dictionary
                user = identity["uid"]
                if "req_user" in kwargs:
                    sids_for_sub = self.sdb.get_by_sub(kwargs["req_user"])
                    if (
                        sids_for_sub
                        and user
                        != self.sdb.get_authentication_event(sids_for_sub[-1]).uid
                    ):
                        logger.debug("Wanted to be someone else!")
                        if "prompt" in areq and "none" in areq["prompt"]:
                            # Need to authenticate but not allowed
                            return redirect_authz_error("login_required", redirect_uri)
                        else:
                            return authn(**authn_args)

        authn_event = AuthnEvent(
            identity["uid"],
            identity.get("salt", ""),
            authn_info=authn_class_ref,
            time_stamp=_ts,
        )

        return {"authn_event": authn_event, "identity": identity, "user": user}

    def setup_session(self, areq, authn_event, cinfo):
        sid = self.sdb.create_authz_session(authn_event, areq)
        self.sdb.do_sub(sid, "")
        return sid

    def authorization_endpoint(self, request="", cookie="", **kwargs):
        """
        Authorize client.

        :param request: The client request
        """
        info = self.auth_init(request)
        if isinstance(info, Response):
            return info

        _cid = info["areq"]["client_id"]
        cinfo = self.cdb[_cid]

        authnres = self.do_auth(
            info["areq"], info["redirect_uri"], cinfo, request, cookie, **kwargs
        )

        if isinstance(authnres, Response):
            return authnres

        logger.debug("- authenticated -")
        logger.debug("AREQ keys: %s" % info["areq"].keys())

        sid = self.setup_session(info["areq"], authnres["authn_event"], cinfo)

        return self.authz_part2(authnres["user"], info["areq"], sid, cookie=cookie)

    def aresp_check(self, aresp, areq):
        return ""

    def create_authn_response(self, areq, sid):
        rtype = areq["response_type"][0]
        _func = self.response_type_map[rtype]
        aresp = _func(
            areq=areq, scode=self.sdb[sid]["code"], sdb=self.sdb, myself=self.baseurl
        )

        if rtype == "code":
            fragment_enc = False
        else:
            fragment_enc = True

        return aresp, fragment_enc

    def response_mode(self, areq, fragment_enc, **kwargs):
        resp_mode = areq["response_mode"]

        if resp_mode == "fragment" and not fragment_enc:
            # Can't be done
            raise InvalidRequest("wrong response_mode")
        elif resp_mode == "query" and fragment_enc:
            # Can't be done
            raise InvalidRequest("wrong response_mode")
        return None

    def authz_part2(self, user, areq, sid, **kwargs):
        """
        After the authentication this is where you should end up.

        :param user:
        :param areq: The Authorization Request
        :param sid: Session key
        :param kwargs: possible other parameters
        :return: A redirect to the redirect_uri of the client
        """
        result = self._complete_authz(user, areq, sid, **kwargs)
        if isinstance(result, Response):
            return result
        else:
            aresp, headers, redirect_uri, fragment_enc = result

        # Mix-Up mitigation
        aresp["iss"] = self.baseurl
        aresp["client_id"] = areq["client_id"]

        # Just do whatever is the default
        location = aresp.request(redirect_uri, fragment_enc)
        logger.debug("Redirected to: '%s' (%s)" % (location, type(location)))
        return SeeOther(str(location), headers=headers)

    def _complete_authz(self, user, areq, sid, **kwargs):
        _log_debug = logger.debug
        _log_debug("- in authenticated() -")

        # Do the authorization
        try:
            permission = self.authz(user, client_id=areq["client_id"])
            self.sdb.update(sid, "permission", permission)
        except Exception:
            raise

        _log_debug("response type: %s" % areq["response_type"])

        if self.sdb.is_revoked(sid):
            return error_response("access_denied", descr="Token is revoked")

        try:
            info = self.create_authn_response(areq, sid)
        except UnSupported as err:
            return error_response(*err.args)

        if isinstance(info, Response):
            return info
        else:
            aresp, fragment_enc = info

        try:
            redirect_uri = self.get_redirect_uri(areq)
        except (RedirectURIError, ParameterError) as err:
            return BadRequest("%s" % err)

        # Must not use HTTP unless implicit grant type and native application

        info = self.aresp_check(aresp, areq)
        if isinstance(info, Response):
            return info

        headers = []
        try:
            _kaka = kwargs["cookie"]
        except KeyError:
            _kaka = None

        c_val = "{}{}{}".format(user, DELIM, areq["client_id"])

        cookie_header = None
        if _kaka is not None:
            if self.cookie_name not in _kaka:  # Don't overwrite
                cookie_header = self.cookie_func(
                    c_val, typ="sso", cookie_name=self.sso_cookie_name, ttl=self.sso_ttl
                )
        else:
            cookie_header = self.cookie_func(
                c_val, typ="sso", cookie_name=self.sso_cookie_name, ttl=self.sso_ttl
            )

        if cookie_header is not None:
            headers.append(cookie_header)
        # Now about the response_mode. Should not be set if it's obvious
        # from the response_type. Knows about 'query', 'fragment' and
        # 'form_post'.

        if "response_mode" in areq:
            try:
                resp = self.response_mode(
                    areq,
                    fragment_enc,
                    aresp=aresp,
                    redirect_uri=redirect_uri,
                    headers=headers,
                )
            except InvalidRequest as err:
                return error_response("invalid_request", str(err))
            else:
                if resp is not None:
                    return resp

        return aresp, headers, redirect_uri, fragment_enc

    def token_scope_check(self, areq, info):
        """Not implemented here."""
        return None

    def token_endpoint(self, request="", authn="", dtype="urlencoded", **kwargs):
        """
        Provide clients with access tokens.

        :param authn: Auhentication info, comes from HTTP header.
        :param request: The request.
        :param dtype: deserialization method for the request.
        """
        logger.debug("- token -")
        logger.debug("token_request: %s" % sanitize(request))

        areq = self.server.message_factory.get_request_type(
            "token_endpoint"
        )().deserialize(request, dtype)

        # Verify client authentication
        try:
            client_id = self.client_authn(self, areq, authn)
        except (FailedAuthentication, AuthnFailure) as err:
            logger.error(err)
            error = TokenErrorResponse(
                error="unauthorized_client", error_description="%s" % err
            )
            return Unauthorized(error.to_json(), content="application/json")

        logger.debug("AccessTokenRequest: %s" % sanitize(areq))

        # `code` is not mandatory for all requests
        if "code" in areq:
            try:
                _info = self.sdb[areq["code"]]
            except KeyError:
                logger.error("Code not present in SessionDB")
                error = TokenErrorResponse(
                    error="unauthorized_client", error_description="Invalid code."
                )
                return Unauthorized(error.to_json(), content="application/json")

            resp = self.token_scope_check(areq, _info)
            if resp:
                return resp
            # If redirect_uri was in the initial authorization request verify that they match
            if (
                "redirect_uri" in _info
                and areq["redirect_uri"] != _info["redirect_uri"]
            ):
                logger.error("Redirect_uri mismatch")
                error = TokenErrorResponse(
                    error="unauthorized_client",
                    error_description="Redirect_uris do not match.",
                )
                return Unauthorized(error.to_json(), content="application/json")
            if "state" in areq:
                if _info["state"] != areq["state"]:
                    logger.error("State value mismatch")
                    error = TokenErrorResponse(
                        error="unauthorized_client",
                        error_description="State values do not match.",
                    )
                    return Unauthorized(error.to_json(), content="application/json")

        # Propagate the client_id further
        areq.setdefault("client_id", client_id)
        grant_type = areq["grant_type"]
        if grant_type == "authorization_code":
            return self.code_grant_type(areq)
        elif grant_type == "refresh_token":
            return self.refresh_token_grant_type(areq)
        elif grant_type == "client_credentials":
            return self.client_credentials_grant_type(areq)
        elif grant_type == "password":
            return self.password_grant_type(areq)
        else:
            raise UnSupported("grant_type: {}".format(grant_type))

    def code_grant_type(self, areq):
        """
        Token authorization using Code Grant.

        RFC6749 section 4.1
        """
        try:
            _tinfo = self.sdb.upgrade_to_token(areq["code"], issue_refresh=True)
        except AccessCodeUsed:
            error = TokenErrorResponse(
                error="invalid_grant", error_description="Access grant used"
            )
            return Unauthorized(error.to_json(), content="application/json")

        logger.debug("_tinfo: %s" % sanitize(_tinfo))

        atr = AccessTokenResponse(**by_schema(AccessTokenResponse, **_tinfo))

        logger.debug("AccessTokenResponse: %s" % sanitize(atr))

        return Response(
            atr.to_json(), content="application/json", headers=OAUTH2_NOCACHE_HEADERS
        )

    def refresh_token_grant_type(self, areq):
        """
        Token refresh.

        RFC6749 section 6
        """
        # This is not implemented here, please see oic.extension.provider.
        return error_response("unsupported_grant_type", descr="Unsupported grant_type")

    def client_credentials_grant_type(self, areq):
        """
        Token authorization using client credentials.

        RFC6749 section 4.4
        """
        # This is not implemented here, please see oic.extension.provider.
        return error_response("unsupported_grant_type", descr="Unsupported grant_type")

    def password_grant_type(self, areq):
        """
        Token authorization using Resource owner password credentials.

        RFC6749 section 4.3
        """
        # This is not implemented here, please see oic.extension.provider.
        return error_response("unsupported_grant_type", descr="Unsupported grant_type")

    def verify_endpoint(self, request="", cookie=None, **kwargs):
        _req = parse_qs(request)
        try:
            areq = parse_qs(_req["query"][0])
        except KeyError:
            return BadRequest("Could not verify endpoint")

        authn, acr = self.pick_auth(areq=areq)
        kwargs["cookie"] = cookie
        return authn.verify(_req, **kwargs)

    def write_session_cookie(self, value, http_only=True, same_site=""):
        return make_cookie(
            self.session_cookie_name,
            value,
            self.seed,
            path="/",
            httponly=http_only,
            same_site=same_site,
        )

    def delete_session_cookie(self):
        return make_cookie(self.session_cookie_name, "", b"", path="/", expire=-1)

    def _compute_session_state(self, state, salt, client_id, redirect_uri):
        parsed_uri = urlparse(redirect_uri)
        rp_origin_url = "{uri.scheme}://{uri.netloc}".format(uri=parsed_uri)

        logger.debug(
            "Calculating sessions state using, client_id:%s origin:%s state:%s salt:%s",
            client_id,
            rp_origin_url,
            state,
            salt,
        )

        session_str = client_id + " " + rp_origin_url + " " + state + " " + salt
        return hashlib.sha256(session_str.encode("utf-8")).hexdigest() + "." + salt



from oic import rndstr
from oic.extension.token import JWTToken

__author__ = "roland"


class NotAllowed(Exception):
    pass


class TokenHandler(object):
    """
    Class for handling tokens.

    Note! the token and refresh token factories both keep their own token databases.
    """

    def __init__(
        self,
        issuer,
        token_policy,
        token_factory=None,
        refresh_token_factory=None,
        keyjar=None,
        sign_alg="RS256",
    ):
        """
        Initialize the class.

        :param token_factory: A callable function that returns a token
        :param refresh_token_factory: A callable function that returns a
            refresh token
        :param token_policy: A dictionary of the form
            {'access_token': {<target_id>: {<grant_type>: <lifetime>}},
             'refresh_token': {<target_id>: {<grant_type>: <lifetime>}}}
        :param keyjar: A oic.utils.keyio.KeyJar instance
        :param sign_alg: Which signature algorithm to use.
        :return: a TokenHandler instance
        """
        self.token_policy = token_policy
        if token_factory is None:
            self.token_factory = JWTToken(
                "T", keyjar=keyjar, iss=issuer, sign_alg=sign_alg
            )
        else:
            self.token_factory = token_factory

        if refresh_token_factory is None:
            self.refresh_token_factory = JWTToken(
                "R",
                keyjar=keyjar,
                iss="https://example.com/as",
                sign_alg=sign_alg,
                token_storage={},
            )
        else:
            self.refresh_token_factory = refresh_token_factory

    def get_access_token(self, target_id, scope, grant_type):
        """
        Return access token for given inputs.

        :param target_id:
        :param scope:
        :param grant_type:
        :return:
        """
        # No default, either there is an explicit policy or there is not
        try:
            lifetime = self.token_policy["access_token"][target_id][grant_type]
        except KeyError:
            raise NotAllowed(
                "Access token for grant_type {} for target_id {} not allowed"
            )

        sid = rndstr(32)
        return self.token_factory(
            sid,
            target_id=target_id,
            scope=scope,
            grant_type=grant_type,
            lifetime=lifetime,
        )

    def refresh_access_token(self, target_id, token, grant_type, **kwargs):
        """
        Return refresh_access_token for given input.

        :param target_id: Who gave me this token
        :param token: The refresh_token
        :param grant_type: Which grant type the token is connected to
        :param kwargs: Extra key word arguments
        :return: New access_token
        """
        # Check that the token is an refresh token
        info = self.refresh_token_factory.get_info(token)

        # Make sure the token should is usable by the client to get a
        # refresh token
        try:
            if target_id != info["azr"]:
                raise NotAllowed("{} can't use this token".format(target_id))
        except KeyError:
            if target_id not in info["aud"]:
                raise NotAllowed("{} can't use this token".format(target_id))

        if self.token_factory.is_valid(info):
            try:
                lifetime = self.token_policy["access_token"][target_id][grant_type]
            except KeyError:
                raise NotAllowed(
                    "Issue access token for grant_type {} for target_id {} not allowed"
                )
            else:
                sid = self.token_factory.db[info["jti"]]
                try:
                    _aud = kwargs["aud"]
                except KeyError:
                    _aud = info["aud"]

                return self.token_factory(
                    sid, target_id=target_id, lifetime=lifetime, aud=_aud
                )

    def get_refresh_token(self, target_id, grant_type, sid):
        try:
            lifetime = self.token_policy["refresh_token"][target_id][grant_type]
        except KeyError:
            raise NotAllowed(
                "Issue access token for grant_type {} for target_id {} not allowed"
            )
        else:
            return self.refresh_token_factory(
                sid, target_id=target_id, lifetime=lifetime
            )

    def invalidate(self, token):
        if self.token_factory.valid(token):
            self.token_factory.invalidate(token)



"""Client management databases."""
from abc import ABCMeta
from abc import abstractmethod
from urllib.parse import quote
from urllib.parse import urljoin

import requests

from oic.oauth2.exception import NoClientInfoReceivedError


class BaseClientDatabase(metaclass=ABCMeta):
    """
    Base implementation for Client management database.

    Custom Client databases should derive from this class.
    They must implement the following methods:
    * ``__getitem__(self, key)``
    * ``__setitem__(self, key, value)``
    * ``__delitem__(self, key)``
    * ``keys(self)``
    * ``items(self)``
    """

    def __init__(self):
        """Perform initialization of storage. Derived classes may override."""

    @abstractmethod
    def __getitem__(self, key):
        """Retrieve an item by a key. Raises KeyError if item not found."""
        pass  # pragma: no cover

    def get(self, key, default=None):
        """Retrieve an item by a key. Return default if not found."""
        try:
            return self[key]
        except KeyError:
            return default

    @abstractmethod
    def __setitem__(self, key, value):
        """Set key with value."""
        pass  # pragma: no cover

    @abstractmethod
    def __delitem__(self, key):
        """Remove key from database."""
        pass  # pragma: no cover

    def __contains__(self, key):
        """Return True if key is contained in the database."""
        try:
            self[key]
        except KeyError:
            return False
        else:
            return True

    @abstractmethod
    def keys(self):
        """Return all contained keys."""
        pass  # pragma: no cover

    @abstractmethod
    def items(self):
        """Return list of all contained items."""
        pass  # pragma: no cover

    def __len__(self):
        """Return number of contained keys."""
        return len(self.keys())


class MDQClient(BaseClientDatabase):
    """Implementation of remote client database."""

    def __init__(self, url):
        """Set the remote storage url."""
        self.url = url
        self.headers = {"Accept": "application/json", "Accept-Encoding": "gzip"}

    def __getitem__(self, item):
        """Retrieve a single entity."""
        mdx_url = urljoin(self.url, "entities/{}".format(quote(item, safe="")))
        response = requests.get(mdx_url, headers=self.headers)
        if response.status_code == 200:
            return response.json()
        else:
            raise NoClientInfoReceivedError(
                "{} {}".format(response.status_code, response.reason)
            )

    def __setitem__(self, item, value):
        """Remote management is readonly."""
        raise RuntimeError("MDQClient is readonly.")

    def __delitem__(self, item):
        """Remote management is readonly."""
        raise RuntimeError("MDQClient is readonly.")

    def keys(self):
        """Get all registered entitites."""
        mdx_url = urljoin(self.url, "entities")
        response = requests.get(mdx_url, headers=self.headers)
        if response.status_code == 200:
            return [item["client_id"] for item in response.json()]
        else:
            raise NoClientInfoReceivedError(
                "{} {}".format(response.status_code, response.reason)
            )

    def items(self):
        """Geting all registered entities."""
        mdx_url = urljoin(self.url, "entities")
        response = requests.get(mdx_url, headers=self.headers)
        if response.status_code == 200:
            return response.json()
        else:
            raise NoClientInfoReceivedError(
                "{} {}".format(response.status_code, response.reason)
            )


# Dictionary can be used as a ClientDatabase
BaseClientDatabase.register(dict)



import inspect
import sys
from urllib.parse import urlparse

import requests

from oic.exception import InvalidRedirectUri
from oic.exception import MissingPage
from oic.exception import RegistrationError
from oic.oauth2.message import OPTIONAL_LIST_OF_SP_SEP_STRINGS
from oic.oauth2.message import OPTIONAL_LIST_OF_STRINGS
from oic.oauth2.message import REQUIRED_LIST_OF_STRINGS
from oic.oauth2.message import SINGLE_OPTIONAL_INT
from oic.oauth2.message import SINGLE_OPTIONAL_STRING
from oic.oauth2.message import SINGLE_REQUIRED_STRING
from oic.oauth2.message import ErrorResponse
from oic.oauth2.message import Message
from oic.oauth2.message import MessageTuple
from oic.oauth2.message import OauthMessageFactory
from oic.oauth2.message import ParamDefinition
from oic.oic.message import JasonWebToken
from oic.utils.http_util import SUCCESSFUL
from oic.utils.jwt import JWT

__author__ = "roland"


# RFC 7662
class TokenIntrospectionRequest(Message):
    c_param = {
        "token": SINGLE_REQUIRED_STRING,
        "token_type_hint": SINGLE_OPTIONAL_STRING,
        # The ones below are part of authentication information
        "client_id": SINGLE_OPTIONAL_STRING,
        "client_assertion_type": SINGLE_OPTIONAL_STRING,
        "client_assertion": SINGLE_OPTIONAL_STRING,
    }


SINGLE_REQUIRED_BOOLEAN = ParamDefinition(bool, True, None, None, False)


class TokenIntrospectionResponse(Message):
    c_param = {
        "active": SINGLE_REQUIRED_BOOLEAN,
        "scope": OPTIONAL_LIST_OF_SP_SEP_STRINGS,
        "client_id": SINGLE_OPTIONAL_STRING,
        "username": SINGLE_OPTIONAL_STRING,
        "token_type": SINGLE_OPTIONAL_STRING,
        "exp": SINGLE_OPTIONAL_INT,
        "iat": SINGLE_OPTIONAL_INT,
        "nbf": SINGLE_OPTIONAL_INT,
        "sub": SINGLE_OPTIONAL_STRING,
        "aud": OPTIONAL_LIST_OF_STRINGS,
        "iss": SINGLE_OPTIONAL_STRING,
        "jti": SINGLE_OPTIONAL_STRING,
    }


# RFC 7009
class TokenRevocationRequest(Message):
    c_param = {
        "token": SINGLE_REQUIRED_STRING,
        "token_type_hint": SINGLE_OPTIONAL_STRING,
        "client_id": SINGLE_OPTIONAL_STRING,
        "client_assertion_type": SINGLE_OPTIONAL_STRING,
        "client_assertion": SINGLE_OPTIONAL_STRING,
    }


class SoftwareStatement(JasonWebToken):
    c_param = JasonWebToken.c_param.copy()
    c_param.update(
        {
            "software_id": SINGLE_OPTIONAL_STRING,
            "client_name": SINGLE_OPTIONAL_STRING,
            "client_uri": SINGLE_OPTIONAL_STRING,
        }
    )


class StateJWT(JasonWebToken):
    c_param = JasonWebToken.c_param.copy()
    c_param.update(
        {
            "aud": SINGLE_OPTIONAL_STRING,
            "rfp": SINGLE_REQUIRED_STRING,
            "kid": SINGLE_OPTIONAL_STRING,
            "target_link__uri": SINGLE_OPTIONAL_STRING,
            "as": SINGLE_OPTIONAL_STRING,
            "at_hash": SINGLE_OPTIONAL_STRING,
            "c_hash": SINGLE_OPTIONAL_STRING,
        }
    )


class RegistrationRequest(Message):
    c_param = {
        "redirect_uris": REQUIRED_LIST_OF_STRINGS,
        "client_name": SINGLE_OPTIONAL_STRING,
        "client_uri": SINGLE_OPTIONAL_STRING,
        "logo_uri": SINGLE_OPTIONAL_STRING,
        "contacts": OPTIONAL_LIST_OF_STRINGS,
        "tos_uri": SINGLE_OPTIONAL_STRING,
        "policy_uri": SINGLE_OPTIONAL_STRING,
        "token_endpoint_auth_method": SINGLE_OPTIONAL_STRING,
        "scope": OPTIONAL_LIST_OF_SP_SEP_STRINGS,
        "grant_types": OPTIONAL_LIST_OF_STRINGS,
        "response_types": OPTIONAL_LIST_OF_STRINGS,
        "jwks_uri": SINGLE_OPTIONAL_STRING,
        "software_id": SINGLE_OPTIONAL_STRING,
        "software_version": SINGLE_OPTIONAL_STRING,
        "software_statement": OPTIONAL_LIST_OF_STRINGS,
    }

    def verify(self, **kwargs):
        if "initiate_login_uri" in self and not self["initiate_login_uri"].startswith(
            "https:"
        ):
            raise RegistrationError("initiate_login_uri is not https")

        if "redirect_uris" in self:
            for uri in self["redirect_uris"]:
                if urlparse(uri).fragment:
                    raise InvalidRedirectUri("redirect_uri contains fragment: %s" % uri)

        for uri in ["client_uri", "logo_uri", "tos_uri", "policy_uri"]:
            if uri in self:
                try:
                    resp = requests.request(
                        "GET", str(self[uri]), allow_redirects=True, verify=False
                    )
                except requests.ConnectionError:
                    raise MissingPage(self[uri])

                if resp.status_code not in SUCCESSFUL:
                    raise MissingPage(self[uri])

        try:
            ss = self["software_statement"]
        except KeyError:
            pass
        else:
            _ss = []
            for _s in ss:
                _ss.append(unpack_software_statement(_s, "", kwargs["keyjar"]))
            self["__software_statement"] = _ss

        return super(RegistrationRequest, self).verify(**kwargs)


class ClientInfoResponse(RegistrationRequest):
    c_param = RegistrationRequest.c_param.copy()
    c_param.update(
        {
            "client_id": SINGLE_REQUIRED_STRING,
            "client_secret": SINGLE_OPTIONAL_STRING,
            "client_id_issued_at": SINGLE_OPTIONAL_INT,
            "client_secret_expires_at": SINGLE_OPTIONAL_INT,
            "registration_access_token": SINGLE_REQUIRED_STRING,
            "registration_client_uri": SINGLE_REQUIRED_STRING,
        }
    )


class ClientRegistrationError(ErrorResponse):
    c_param = ErrorResponse.c_param.copy()
    c_param.update({"state": SINGLE_OPTIONAL_STRING})
    c_allowed_values = ErrorResponse.c_allowed_values.copy()
    c_allowed_values.update(
        {
            "error": [
                "invalid_redirect_uri",
                "invalid_client_metadata",
                "invalid_client_id",
            ]
        }
    )


class ClientUpdateRequest(RegistrationRequest):
    c_param = RegistrationRequest.c_param.copy()
    c_param.update(
        {
            "client_id": SINGLE_REQUIRED_STRING,
            "client_secret": SINGLE_OPTIONAL_STRING,
            "client_assertion_type": SINGLE_OPTIONAL_STRING,
            "client_assertion": SINGLE_OPTIONAL_STRING,
        }
    )


MSG = {
    "RegistrationRequest": RegistrationRequest,
    "ClientInfoResponse": ClientInfoResponse,
    "ClientRegistrationError": ClientRegistrationError,
    "ClientUpdateRequest": ClientUpdateRequest,
    "TokenRevocationRequest": TokenRevocationRequest,
    "TokenIntrospectionRequest": TokenIntrospectionRequest,
    "TokenIntrospectionResponse": TokenIntrospectionResponse,
    "SoftwareStatement": SoftwareStatement,
    "StateJWT": StateJWT,
}


def factory(msgtype):
    for _, obj in inspect.getmembers(sys.modules[__name__]):
        if inspect.isclass(obj) and issubclass(obj, Message):
            try:
                if obj.__name__ == msgtype:
                    return obj
            except AttributeError:
                pass

    # check among standard OAuth2 messages
    from oic.oauth2 import message

    return message.factory(msgtype)


def make_software_statement(keyjar, iss, **kwargs):
    params = list(inspect.signature(JWT.__init__).parameters.keys())

    params.remove("self")

    args = {}
    for param in params:
        try:
            args[param] = kwargs[param]
        except KeyError:
            pass
        else:
            del kwargs[param]

    _jwt = JWT(keyjar, msgtype=SoftwareStatement, iss=iss, **args)
    return _jwt.pack(**kwargs)


def unpack_software_statement(software_statement, iss, keyjar):
    _jwt = JWT(keyjar, iss=iss, msgtype=SoftwareStatement)
    return _jwt.unpack(software_statement)


class ExtensionMessageFactory(OauthMessageFactory):
    """Message factory for Extension code."""

    introspection_endpoint = MessageTuple(
        TokenIntrospectionRequest, TokenIntrospectionResponse
    )
    revocation_endpoint = MessageTuple(TokenRevocationRequest, Message)
    registration_endpoint = MessageTuple(RegistrationRequest, ClientInfoResponse)
    update_endpoint = MessageTuple(ClientUpdateRequest, ClientInfoResponse)
    delete_endpoint = MessageTuple(ClientUpdateRequest, ClientInfoResponse)



import logging
from http import cookiejar as http_cookiejar
from http.cookiejar import http2time  # type: ignore
from typing import Any
from typing import Dict
from typing import Optional
from urllib.parse import parse_qs
from urllib.parse import urlsplit
from urllib.parse import urlunsplit

from typing_extensions import Literal

from oic.exception import UnSupported
from oic.oauth2.exception import TimeFormatError
from oic.utils.sanitize import sanitize

logger = logging.getLogger(__name__)

__author__ = "roland"

URL_ENCODED = "application/x-www-form-urlencoded"
JSON_ENCODED = "application/json"

DEFAULT_POST_CONTENT_TYPE = URL_ENCODED

PAIRS = {
    "port": "port_specified",
    "domain": "domain_specified",
    "path": "path_specified",
}

ATTRS: Dict[str, Any] = {
    "version": None,
    "name": "",
    "value": None,
    "port": None,
    "port_specified": False,
    "domain": "",
    "domain_specified": False,
    "domain_initial_dot": False,
    "path": "",
    "path_specified": False,
    "secure": False,
    "expires": None,
    "discard": True,
    "comment": None,
    "comment_url": None,
    "rest": "",
    "rfc2109": True,
}

# The encodings understood by our Message class
ENCODINGS = Literal["json", "urlencoded", "dict", "jwt", "jwe"]


def get_or_post(
    uri, method, req, content_type=DEFAULT_POST_CONTENT_TYPE, accept=None, **kwargs
):
    """
    Construct HTTP request.

    :param uri:
    :param method:
    :param req:
    :param content_type:
    :param accept:
    :param kwargs:
    :return:
    """
    if method in ["GET", "DELETE"]:
        if req.keys():
            _req = req.copy()
            comp = urlsplit(str(uri))
            if comp.query:
                _req.update(parse_qs(comp.query))

            _query = str(_req.to_urlencoded())
            path = urlunsplit(
                (comp.scheme, comp.netloc, comp.path, _query, comp.fragment)
            )
        else:
            path = uri
        body = None
    elif method in ["POST", "PUT"]:
        path = uri
        if content_type == URL_ENCODED:
            body = req.to_urlencoded()
        elif content_type == JSON_ENCODED:
            body = req.to_json()
        else:
            raise UnSupported("Unsupported content type: '%s'" % content_type)

        header_ext = {"Content-Type": content_type}
        if accept:
            header_ext = {"Accept": accept}

        if "headers" in kwargs.keys():
            kwargs["headers"].update(header_ext)
        else:
            kwargs["headers"] = header_ext
    else:
        raise UnSupported("Unsupported HTTP method: '%s'" % method)

    return path, body, kwargs


def set_cookie(cookiejar, kaka):
    """
    Place a cookie (a http_cookielib.Cookie based on a set-cookie header line) in the cookie jar.

    Always chose the shortest expires time.

    :param cookiejar:
    :param kaka: Cookie
    """
    # default rfc2109=False
    # max-age, httponly
    for cookie_name, morsel in kaka.items():
        std_attr = ATTRS.copy()
        std_attr["name"] = cookie_name
        _tmp = morsel.coded_value
        if _tmp.startswith('"') and _tmp.endswith('"'):
            std_attr["value"] = _tmp[1:-1]
        else:
            std_attr["value"] = _tmp

        std_attr["version"] = 0
        attr = ""
        # copy attributes that have values
        try:
            for attr in morsel.keys():
                if attr in ATTRS:
                    if morsel[attr]:
                        if attr == "expires":
                            std_attr[attr] = http2time(morsel[attr])
                        else:
                            std_attr[attr] = morsel[attr]
                elif attr == "max-age":
                    if morsel[attr]:
                        std_attr["expires"] = http2time(morsel[attr])
        except TimeFormatError:
            # Ignore cookie
            logger.info(
                "Time format error on %s parameter in received cookie"
                % (sanitize(attr),)
            )
            continue

        for att, spec in PAIRS.items():
            if std_attr[att]:
                std_attr[spec] = True

        if std_attr["domain"] and std_attr["domain"].startswith("."):
            std_attr["domain_initial_dot"] = True

        if morsel["max-age"] == 0:
            try:
                cookiejar.clear(
                    domain=std_attr["domain"],
                    path=std_attr["path"],
                    name=std_attr["name"],
                )
            except ValueError:
                pass
        else:
            # Fix for Microsoft cookie error
            if "version" in std_attr:
                try:
                    std_attr["version"] = std_attr["version"].split(",")[0]
                except (TypeError, AttributeError):
                    pass

            new_cookie = http_cookiejar.Cookie(**std_attr)  # type: ignore

            cookiejar.set_cookie(new_cookie)


def match_to_(val, vlist):
    if isinstance(vlist, str):
        if vlist.startswith(val):
            return True
    else:
        for v in vlist:
            if v.startswith(val):
                return True
    return False


def guess_body_type(reqresp):
    """Try to guess the body type of the message from a response object."""
    # try to handle chunked responses.
    if "chunked" not in reqresp.headers.get("transfer-encoding", ""):
        if int(reqresp.headers["content-length"]) == 0:
            return None

    _ctype = reqresp.headers["content-type"]
    if match_to_("application/json", _ctype):
        body_type = "json"
    elif match_to_("application/jwt", _ctype):
        body_type = "jwt"
    elif match_to_("application/jwe", _ctype):
        body_type = "jwe"
    elif match_to_(URL_ENCODED, _ctype):
        body_type = "urlencoded"
    else:
        body_type = None
    return body_type


def verify_header(reqresp, body_type: Optional[ENCODINGS]) -> Optional[ENCODINGS]:
    logger.debug("resp.headers: %s" % (sanitize(reqresp.headers),))
    logger.debug("resp.txt: %s" % (sanitize(reqresp.text),))

    if body_type is None:
        return guess_body_type(reqresp)

    if body_type == "json":
        if not match_to_("application/json", reqresp.headers["content-type"]):
            if match_to_("application/jwt", reqresp.headers["content-type"]):
                body_type = "jwt"
            else:
                raise ValueError(
                    "content-type: %s" % (reqresp.headers["content-type"],)
                )
    elif body_type == "jwt":
        if not match_to_("application/jwt", reqresp.headers["content-type"]):
            raise ValueError(
                "Wrong content-type in header, got: {} expected "
                "'application/jwt'".format(reqresp.headers["content-type"])
            )
    elif body_type == "urlencoded":
        if not match_to_(DEFAULT_POST_CONTENT_TYPE, reqresp.headers["content-type"]):
            if not match_to_("text/plain", reqresp.headers["content-type"]):
                raise ValueError("Wrong content-type")
    else:
        raise ValueError("Unknown return format: %s" % body_type)

    return body_type



from oic.exception import PyoidcError

__author__ = "roland"


class HttpError(PyoidcError):
    pass


class MissingRequiredAttribute(PyoidcError):
    pass


class VerificationError(PyoidcError):
    pass


class ResponseError(PyoidcError):
    pass


class TimeFormatError(PyoidcError):
    pass


class CapabilitiesMisMatch(PyoidcError):
    pass


class MissingEndpoint(PyoidcError):
    pass


class TokenError(PyoidcError):
    pass


class GrantError(PyoidcError):
    pass


class ParseError(PyoidcError):
    pass


class OtherError(PyoidcError):
    pass


class NoClientInfoReceivedError(PyoidcError):
    pass


class InvalidRequest(PyoidcError):
    pass


class NonFatalException(PyoidcError):
    """
    Return the response but accompany it by an error message.

    :param resp: A response that the function/method would return on non-error
    :param msg: A message describing what error has occurred.
    """

    def __init__(self, resp, msg):
        self.resp = resp
        self.msg = msg


class Unsupported(PyoidcError):
    pass


class UnsupportedResponseType(Unsupported):
    pass



import json
import time
from abc import ABCMeta
from abc import abstractmethod
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Union
from typing import cast

from oic.utils.time_util import time_sans_frac


class AuthnEvent(object):
    def __init__(
        self,
        uid,
        salt,
        valid=3600,
        authn_info=None,
        time_stamp=0,
        authn_time=None,
        valid_until=None,
    ):
        """
        Create a representation of an authentication event.

        :param uid: The local user identifier
        :param salt: Salt to be used in creating a sub
        :param valid: How long the authentication is expected to be valid
        :param authn_info: Info about the authentication event
        :return:
        """
        self.uid = uid
        self.salt = salt
        self.authn_time = authn_time or (int(time_stamp) or time_sans_frac())
        self.valid_until = valid_until or (self.authn_time + int(valid))
        self.authn_info = authn_info

    def valid(self):
        return self.valid_until > time.time()

    def valid_for(self):
        return self.valid_until - time.time()

    def to_json(self):
        """Serialize AuthnEvent to JSON."""
        return json.dumps(self.__dict__)

    @classmethod
    def from_json(cls, json_struct):
        """Create AuthnEvent from JSON."""
        dic = json.loads(json_struct)
        return cls(**dic)


class SessionBackend(metaclass=ABCMeta):
    """Backend for storing sessionDB data."""

    @abstractmethod
    def __setitem__(self, key: str, value: Dict[str, Union[str, bool]]) -> None:
        """Store the session information under the session_id."""

    @abstractmethod
    def __getitem__(self, key: str) -> Dict[str, Union[str, bool]]:
        """
        Retrieve the session information based os session_id.

        @raises KeyError when no key is found.
        """

    @abstractmethod
    def __delitem__(self, key: str) -> None:
        """Remove the stored session from storage."""

    @abstractmethod
    def __contains__(self, key: str) -> bool:
        """Test presence of key in storage."""

    @abstractmethod
    def get_by_uid(self, uid: str) -> List[str]:
        """Return session ids (keys) based on `uid` (internal user identifier)."""

    @abstractmethod
    def get_by_sub(self, sub: str) -> List[str]:
        """Return session ids based on `sub` (external user identifier)."""

    @abstractmethod
    def get(self, attr: str, val: str) -> List[str]:
        """Return session ids based on attribute name and value."""

    def get_client_ids_for_uid(self, uid: str) -> List[str]:
        """Return client ids that have a session for given uid."""
        return [cast(str, self[sid]["client_id"]) for sid in self.get_by_uid(uid)]

    def get_verified_logout(self, uid: str) -> Optional[str]:
        """Return logout verification key for given uid."""
        # Since all the sessions should be the same, we return the first one
        sids = self.get_by_uid(uid)
        if len(sids) == 0:
            return None
        _dict = self[sids[0]]
        if "verified_logout" not in _dict:
            return None
        return cast(str, _dict["verified_logout"])

    def get_token_ids(self, uid: str) -> List[str]:
        """Return id_tokens for the given uid."""
        return [cast(str, self[sid]["id_token"]) for sid in self.get_by_uid(uid)]

    def is_revoke_uid(self, uid: str) -> bool:
        """Return if the session is revoked."""
        # We do not care which session it is - once revoked, al are revoked
        return any([self[sid]["revoked"] for sid in self.get_by_uid(uid)])

    def update(self, key: str, attribute: str, value: Any):
        """
        Update information stored. If the key is not know a new entry will be constructed.

        :param key: Key to the database
        :param attribute: Attribute name
        :param value: Attribute value
        """
        if key not in self:
            self[key] = {attribute: value}
        else:
            item = self[key]
            item[attribute] = value
            self[key] = item

    def get_uid_by_sub(self, sub: str) -> Optional[str]:
        """Return User id based on sub."""
        for sid in self.get_by_sub(sub):
            return AuthnEvent.from_json(self[sid]["authn_event"]).uid
        return None

    def get_uid_by_sid(self, sid: str) -> str:
        """Return User id based on session ID."""
        return AuthnEvent.from_json(self[sid]["authn_event"]).uid


class DictSessionBackend(SessionBackend):
    """
    Simple implementation of `SessionBackend` based on dictionary.

    This should really not be used in production.
    """

    def __init__(self):
        """Create the storage."""
        self.storage: Dict[str, Dict[str, Union[str, bool]]] = {}

    def __setitem__(self, key: str, value: Dict[str, Union[str, bool]]) -> None:
        """Store the session info in the storage."""
        self.storage[key] = value

    def __getitem__(self, key: str) -> Dict[str, Union[str, bool]]:
        """Retrieve session information based on session id."""
        return self.storage[key]

    def __delitem__(self, key: str) -> None:
        """Delete the session info."""
        del self.storage[key]

    def __contains__(self, key: str) -> bool:
        return key in self.storage

    def get_by_sub(self, sub: str) -> List[str]:
        """Return session ids based on sub."""
        return [
            sid for sid, session in self.storage.items() if session.get("sub") == sub
        ]

    def get_by_uid(self, uid: str) -> List[str]:
        """Return session ids based on uid."""
        return [
            sid
            for sid, session in self.storage.items()
            if AuthnEvent.from_json(session["authn_event"]).uid == uid
        ]

    def get(self, attr: str, val: str) -> List[str]:
        """Return session ids based on attribute name and value."""
        return [
            sid for sid, session in self.storage.items() if session.get(attr) == val
        ]



import copy
import hashlib
import logging
from typing import Dict
from typing import Type
from typing import Union
from typing import cast
from urllib.parse import urlsplit

from oic import rndstr
from oic.extension import client
from oic.extension.message import ClientInfoResponse
from oic.oauth2 import AccessTokenResponse
from oic.oauth2 import AuthorizationRequest
from oic.oauth2 import AuthorizationResponse
from oic.oauth2 import ErrorResponse
from oic.oauth2 import ResponseError
from oic.oauth2 import TokenError
from oic.oauth2.message import ASConfigurationResponse
from oic.utils.authn.client import CLIENT_AUTHN_METHOD
from oic.utils.http_util import Redirect
from oic.utils.sanitize import sanitize
from oic.utils.webfinger import WebFinger

__author__ = "roland"


logger = logging.getLogger(__name__)


class OAuth2Error(Exception):
    pass


class OAuthClient(client.Client):
    def __init__(
        self,
        client_id=None,
        client_prefs=None,
        client_authn_method=None,
        keyjar=None,
        verify_ssl=True,
        behaviour=None,
        jwks_uri="",
        kid=None,
    ):
        client.Client.__init__(
            self, client_id, client_authn_method, keyjar=keyjar, verify_ssl=verify_ssl
        )
        self.behaviour = behaviour or {}
        self.userinfo_request_method = ""
        self.allow_sign_alg_none = False
        self.authz_req = {}
        self.get_userinfo = True
        self.jwks_uri = jwks_uri
        self.kid = kid
        self.client_prefs = client_prefs
        # Make it the same. Been bitten by this too many times !
        self.keyjar.verify_ssl = verify_ssl

    def create_authn_request(self, session, acr_value=None, **kwargs):
        assert self.registration_response is not None  # nosec
        session["state"] = rndstr(32)
        request_args = {
            "response_type": self.behaviour["response_type"],
            "state": session["state"],
            "redirect_uri": self.registration_response["redirect_uris"][0],
        }

        try:
            request_args["scope"] = self.behaviour["scope"]
        except KeyError:
            pass

        request_args.update(kwargs)
        cis = self.construct_AuthorizationRequest(request_args=request_args)
        logger.debug("request: %s" % sanitize(cis))

        url, body, ht_args, cis = cast(
            AuthorizationRequest,
            self.uri_and_body(
                AuthorizationRequest, cis, method="GET", request_args=request_args
            ),
        )

        self.authz_req[request_args["state"]] = cis
        logger.debug("body: %s" % sanitize(body))
        logger.info("URL: %s" % sanitize(url))
        logger.debug("ht_args: %s" % sanitize(ht_args))

        resp = Redirect(str(url))
        if ht_args:
            resp.headers.extend([(a, b) for a, b in ht_args.items()])
        logger.debug("resp_headers: %s" % sanitize(resp.headers))
        return resp

    def has_access_token(self, **kwargs):
        try:
            token = self.get_token(**kwargs)
        except TokenError:
            pass
        else:
            if token.access_token:
                return True
        return False

    def _err(self, txt):
        logger.error(sanitize(txt))
        raise OAuth2Error(txt)

    def callback(self, response, session, format="dict"):
        """
        Call when an AuthN response has been received from the OP.

        :param response: The URL returned by the OP
        :return:
        """
        if self.behaviour["response_type"] == "code":
            respcls: Union[
                Type[AuthorizationResponse], Type[AccessTokenResponse]
            ] = AuthorizationResponse
        else:
            respcls = AccessTokenResponse

        try:
            authresp = self.parse_response(
                respcls, response, sformat=format, keyjar=self.keyjar
            )
        except ResponseError:
            msg = "Could not parse response: '{}'"
            logger.error(msg.format(sanitize(response)))
            raise OAuth2Error("Problem parsing response")

        logger.info("{}: {}".format(respcls.__name__, sanitize(authresp)))

        if isinstance(authresp, ErrorResponse):
            if authresp["error"] == "login_required":
                return self.create_authn_request(session)
            else:
                raise OAuth2Error("Access denied")

        if authresp["state"] not in self.authz_req:
            self._err("Received state not the same as expected.")

        if self.behaviour["response_type"] == "code":
            # get the access token
            assert self.registration_response is not None  # nosec
            try:
                args = {
                    "code": authresp["code"],
                    "redirect_uri": self.registration_response["redirect_uris"][
                        0
                    ],  # type: ignore
                    "client_id": self.client_id,
                    "client_secret": self.client_secret,
                }

                try:
                    args["scope"] = response["scope"]
                except KeyError:
                    pass

                atresp = self.do_access_token_request(
                    state=authresp["state"],
                    request_args=args,
                    authn_method=self.registration_response[
                        "token_endpoint_auth_method"
                    ],
                )
                logger.info("Access token response: {}".format(sanitize(atresp)))
            except Exception as err:
                logger.error("%s" % err)
                raise

            if isinstance(atresp, ErrorResponse):
                self._err("Error response: {}".format(atresp.to_dict()))

            _token = atresp["access_token"]
        else:
            _token = authresp["access_token"]

        return {"access_token": _token}


class OAuthClients(object):
    def __init__(self, config, base_url, seed="", jwks_info=None, verify_ssl=True):
        """
        Initialize the client.

        :param config: Imported configuration module
        :return:
        """
        self.client: Dict[str, OAuthClient] = {}
        self.client_cls = OAuthClient
        self.config = config
        self.seed = seed or rndstr(16)
        self.seed = self.seed.encode("utf8")
        self.path: Dict[str, str] = {}
        self.base_url = base_url
        self.jwks_info = jwks_info
        self.verify_ssl = verify_ssl

        for key, val in config.CLIENTS.items():
            if self.jwks_info:
                _val = val.copy()
                _val.update(self.jwks_info)
            else:
                _val = val

            if key == "":
                continue
            else:
                self.client[key] = self.create_client(**_val)

    def get_path(self, redirect_uris, issuer):
        for ruri in redirect_uris:
            p = urlsplit(ruri)
            self.path[p.path[1:]] = issuer

    def create_client(self, **kwargs):
        """
        Do an instantiation of a client instance.

        :param: Keyword arguments
            Keys are:
                srv_discovery_url
                client_info
                client_registration
                provider_info
                behaviour
        :return: client instance
        """
        _key_set = set(list(kwargs.keys()))
        try:
            _verify_ssl = kwargs["verify_ssl"]
        except KeyError:
            _verify_ssl = self.verify_ssl
        else:
            _key_set.discard("verify_ssl")

        _client = self.client_cls(
            client_authn_method=CLIENT_AUTHN_METHOD,
            behaviour=kwargs["behaviour"],
            verify_ssl=_verify_ssl,
        )

        # The behaviour parameter is not significant for the election process
        _key_set.discard("behaviour")

        for param in ["allow"]:
            try:
                setattr(_client, param, kwargs[param])
            except KeyError:
                pass
            else:
                _key_set.discard(param)

        if _key_set == {"client_info", "srv_discovery_url"}:
            # Ship the webfinger part
            # Gather OP information
            _client.provider_config(kwargs["srv_discovery_url"])
            # register the client
            _client.register(
                _client.provider_info["registration_endpoint"], **kwargs["client_info"]
            )
            self.get_path(
                kwargs["client_info"]["redirect_uris"], kwargs["srv_discovery_url"]
            )
        elif _key_set == {"provider_info", "client_info"}:
            _client.handle_provider_config(
                ASConfigurationResponse(**kwargs["provider_info"]),
                kwargs["provider_info"]["issuer"],
            )
            _client.register(
                _client.provider_info["registration_endpoint"], **kwargs["client_info"]
            )

            self.get_path(
                kwargs["client_info"]["redirect_uris"],
                kwargs["provider_info"]["issuer"],
            )
        elif _key_set == {"provider_info", "client_registration"}:
            _client.handle_provider_config(
                ASConfigurationResponse(**kwargs["provider_info"]),
                kwargs["provider_info"]["issuer"],
            )
            _client.store_registration_info(
                ClientInfoResponse(**kwargs["client_registration"])
            )
            self.get_path(
                kwargs["client_info"]["redirect_uris"],
                kwargs["provider_info"]["issuer"],
            )
        elif _key_set == {"srv_discovery_url", "client_registration"}:
            _client.provider_config(kwargs["srv_discovery_url"])
            _client.store_registration_info(
                ClientInfoResponse(**kwargs["client_registration"])
            )
            self.get_path(
                kwargs["client_registration"]["redirect_uris"],
                kwargs["srv_discovery_url"],
            )
        else:
            raise Exception("Configuration error ?")

        return client

    def dynamic_client(self, issuer="", userid=""):
        client = self.client_cls(
            client_authn_method=CLIENT_AUTHN_METHOD,
            verify_ssl=self.verify_ssl,
            **self.jwks_info,
        )
        if userid:
            wf = WebFinger(httpd=client)
            issuer = wf.discovery_query(userid)

        if not issuer:
            raise OAuth2Error("Missing issuer")

        logger.info("issuer: {}".format(issuer))

        if issuer in self.client:
            return self.client[issuer]
        else:
            # Gather OP information
            _pcr = client.provider_config(issuer)
            logger.info("Provider info: {}".format(sanitize(_pcr.to_dict())))
            issuer = _pcr["issuer"]  # So no hickup later about trailing '/'
            # register the client
            _cinfo = self.config.CLIENTS[""]["client_info"]
            reg_args = copy.copy(_cinfo)
            h = hashlib.sha256(self.seed)
            h.update(issuer.encode("utf8"))  # issuer has to be bytes
            base_urls = _cinfo["redirect_uris"]

            reg_args["redirect_uris"] = [
                u.format(base=self.base_url, iss=h.hexdigest()) for u in base_urls
            ]
            try:
                reg_args["post_logout_redirect_uris"] = [
                    u.format(base=self.base_url, iss=h.hexdigest())
                    for u in reg_args["post_logout_redirect_uris"]
                ]
            except KeyError:
                pass

            self.get_path(reg_args["redirect_uris"], issuer)
            if client.jwks_uri:
                reg_args["jwks_uri"] = client.jwks_uri

            rr = client.register(_pcr["registration_endpoint"], **reg_args)
            msg = "Registration response: {}"
            logger.info(msg.format(sanitize(rr.to_dict())))

            try:
                client.behaviour.update(**self.config.CLIENTS[""]["behaviour"])
            except KeyError:
                pass

            self.client[issuer] = client
            return client

    def __getitem__(self, item):
        """
        Given a service identifier return a suitable client.

        :param item:
        :return:
        """
        try:
            return self.client[item]
        except KeyError:
            return self.dynamic_client(issuer=item)

    def __delitem__(self, key):
        del self.client[key]

    def keys(self):
        return list(self.client.keys())

    def return_paths(self):
        return self.path.keys()



import hashlib
import json
from base64 import urlsafe_b64encode

from jwkest import JWKESTException
from jwkest import jws
from jwkest.jws import JWS


class ValidationError(Exception):
    pass


def get_hash_size(alg):
    return int(alg[2:])


def hash_value(size, data):
    data = data.encode("utf-8")
    if size == 256:
        return hashlib.sha256(data).digest()
    elif size == 384:
        return hashlib.sha384(data).digest()
    elif size == 512:
        return hashlib.sha512(data).digest()

    raise ValueError("The specifed hash size '{}' is unsupported.".format(size))


def serialize_dict(data, serialization_template):
    buffer = []
    keys = []
    for key in data:
        keys.append(key)
        buffer.append(serialization_template.format(key, data[key]))

    return keys, "".join(buffer)


def b64_hash(val, hash_size):
    return urlsafe_b64encode(hash_value(hash_size, val)).decode("utf-8")


def _equals(value, expected):
    if value != expected:
        raise ValidationError("{} != {}".format(value, expected))


def _serialize_params(params, str_format, hash_size):
    _keys, _buffer = serialize_dict(params, str_format)
    _hash = b64_hash(_buffer, hash_size)
    return [_keys, _hash]


def _verify_params(params, req, str_format, hash_size, strict_verification, key):
    key_order, req_hash = req

    if strict_verification and len(key_order) != len(params):
        raise ValidationError("Too many or too few {}".format(key))

    buffer = ""
    try:
        buffer = "".join([str_format.format(k, params[k]) for k in key_order])
    except KeyError:
        raise ValidationError("Too few {}".format(key))

    _equals(req_hash, b64_hash(buffer, hash_size))


def _upper(s):
    return s.upper()


SIMPLE_OPER = {
    "method": ("m", _upper),
    "host": ("u", None),
    "path": ("p", None),
    "time_stamp": ("ts", int),
}

QUERY_PARAM_FORMAT = "{}={}"
REQUEST_HEADER_FORMAT = "{}: {}"

PARAM_ARGS = {
    "query_params": ("q", QUERY_PARAM_FORMAT),
    "headers": ("h", REQUEST_HEADER_FORMAT),
}


class SignedHttpRequest(object):
    def __init__(self, key):
        self.key = key

    def sign(self, alg, **kwargs):
        http_json = {}
        hash_size = get_hash_size(alg)

        for arg, (key, func) in SIMPLE_OPER.items():
            try:
                if func is None:
                    http_json[key] = kwargs[arg]
                else:
                    http_json[key] = func(kwargs[arg])
            except KeyError:
                pass

        for arg, (key, format) in PARAM_ARGS.items():
            try:
                http_json[key] = _serialize_params(kwargs[arg], format, hash_size)
            except KeyError:
                pass

        try:
            http_json["b"] = b64_hash(kwargs["body"], hash_size)
        except KeyError:
            pass

        if not http_json:
            raise ValueError("No data to sign")

        jws = JWS(json.dumps(http_json), alg=alg, typ="pop")
        _jwt = jws.sign_compact(keys=[self.key])
        return _jwt

    def verify(self, signature, **kwargs):
        _jw = jws.factory(signature)
        if not _jw:
            raise ValidationError("Not a signed request")

        try:
            unpacked_req = _jw.verify_compact(signature, keys=[self.key])
        except JWKESTException:
            raise ValidationError("Could not verify signature")

        _header = _jw.jwt.headers
        if "typ" not in _header or _header["typ"] != "pop":
            raise ValidationError("Incorrect JWS header 'typ', must be 'pop'")

        hash_size = get_hash_size(_header["alg"])

        for arg, (key, func) in SIMPLE_OPER.items():
            if arg == "time_stamp":
                continue
            try:
                if func is None:
                    _val = kwargs[arg]
                else:
                    _val = func(kwargs[arg])
                _equals(unpacked_req[key], _val)
            except KeyError:
                pass

        for arg, (key, format) in PARAM_ARGS.items():
            try:
                _attr = "strict_{}_verification".format(arg)
                _strict_verify = kwargs[_attr]
            except KeyError:
                _strict_verify = False

            try:
                _verify_params(
                    kwargs[arg],
                    unpacked_req[key],
                    format,
                    hash_size,
                    _strict_verify,
                    key,
                )
            except KeyError:
                pass

        if "b" not in unpacked_req and "body" not in kwargs:
            pass
        elif "b" in unpacked_req and "body" in kwargs:
            _equals(
                b64_hash(kwargs.get("body", ""), hash_size), unpacked_req.get("b", "")
            )
        else:
            if "b" in unpacked_req:
                raise ValidationError("Body sent but not received!!")
            else:
                raise ValidationError("Body received but not sent!!")

        return unpacked_req



#!/usr/bin/env python
#
# Copyright (C) 2013 Umea Universitet, Sweden
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import re
import sys
from io import open

from setuptools import setup
from setuptools.command.test import test as TestCommand

__author__ = 'rohe0002'


class PyTest(TestCommand):
    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        # import here, cause outside the eggs aren't loaded
        import pytest

        errno = pytest.main(self.test_args)
        sys.exit(errno)


tests_requires = ['responses', 'testfixtures', 'pytest', 'freezegun']

with open('src/oic/__init__.py', 'r') as fd:
    version = re.search(r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]',
                        fd.read(), re.MULTILINE).group(1)

setup(
    name="oic",
    version=version,
    description="Python implementation of OAuth2 and OpenID Connect",
    long_description=open('README.rst', encoding='utf-8').read(),
    author="Roland Hedberg",
    author_email="roland@catalogix.se",
    license="Apache 2.0",
    url='https://github.com/OpenIDC/pyoidc/',
    packages=[
        "oic", "oic/oauth2", "oic/oic", "oic/utils", "oic/utils/authn",
        "oic/utils/userinfo", 'oic/utils/rp', 'oic/extension'
    ],
    entry_points={
        'console_scripts': [
            'oic-client-management = oic.utils.client_management:run'
        ]
    },
    package_dir={"": "src"},
    package_data={"oic": ["py.typed"]},
    include_package_data=True,
    classifiers=[
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Topic :: Software Development :: Libraries :: Python Modules"],
    python_requires='~=3.7',
    extras_require={
        'develop': ["cherrypy==3.2.4", "pyOpenSSL"],
        'testing': tests_requires,
        'docs': ['Sphinx', 'sphinx-autobuild', 'alabaster'],
        'quality': ['pylama', 'isort', 'eradicate', 'mypy', 'black', 'bandit', 'readme_renderer[md]'],
        'types': ['types-requests'],
        'ldap_authn': ['python-ldap'],
        'examples': ['beaker'],
    },
    install_requires=[
        "requests",
        "pycryptodomex",
        "pyjwkest>=1.3.6",
        "mako",
        "cryptography",
        "defusedxml",
        "typing_extensions",
    ],
    tests_require=tests_requires,
    long_description_content_type="text/x-rst",
    zip_safe=False,
    cmdclass={'test': PyTest},
)



import copy
import hashlib
import logging
from typing import Dict
from urllib.parse import urlsplit

from oic import oic
from oic import rndstr
from oic.exception import MissingAttribute
from oic.oauth2 import ErrorResponse
from oic.oauth2 import ResponseError
from oic.oauth2 import TokenError
from oic.oauth2.message import ASConfigurationResponse
from oic.oic import AuthorizationRequest
from oic.oic import AuthorizationResponse
from oic.oic import RegistrationResponse
from oic.oic.message import OpenIDSchema
from oic.utils.authn.client import CLIENT_AUTHN_METHOD
from oic.utils.http_util import Redirect
from oic.utils.sanitize import sanitize

__author__ = "roland"


logger = logging.getLogger(__name__)


class OIDCError(Exception):
    pass


class Client(oic.Client):
    def __init__(
        self,
        client_id=None,
        client_prefs=None,
        client_authn_method=None,
        keyjar=None,
        verify_ssl=True,
        behaviour=None,
        config=None,
        jwks_uri="",
        kid=None,
    ):
        oic.Client.__init__(
            self,
            client_id,
            client_prefs,
            client_authn_method,
            keyjar,
            verify_ssl,
            config=config,
        )
        if behaviour:
            self.behaviour = behaviour
        self.userinfo_request_method = ""
        self.allow_sign_alg_none = False
        self.authz_req = {}
        self.get_userinfo = True
        self.oidc = True
        self.jwks_uri = jwks_uri
        self.kid = kid

    def create_authn_request(self, session, acr_value=None, **kwargs):
        session["state"] = rndstr(32)
        request_args = {
            "response_type": self.behaviour["response_type"],
            "scope": self.behaviour["scope"],
            "state": session["state"],
            "redirect_uri": self.registration_response["redirect_uris"][0],
        }

        if self.oidc:
            session["nonce"] = rndstr(32)
            request_args["nonce"] = session["nonce"]

        if acr_value is not None:
            request_args["acr_values"] = acr_value

        request_args.update(kwargs)
        cis = self.construct_AuthorizationRequest(request_args=request_args)
        logger.debug("request: %s" % sanitize(cis))

        url, body, ht_args, cis = self.uri_and_body(
            AuthorizationRequest, cis, method="GET", request_args=request_args
        )

        self.authz_req[request_args["state"]] = cis
        logger.debug("body: %s" % sanitize(body))
        logger.info("URL: %s" % sanitize(url))
        logger.debug("ht_args: %s" % sanitize(ht_args))

        resp = Redirect(str(url))
        if ht_args:
            resp.headers.extend([(a, b) for a, b in ht_args.items()])
        logger.debug("resp_headers: %s" % sanitize(resp.headers))
        return resp

    def has_access_token(self, **kwargs):
        try:
            token = self.get_token(**kwargs)
        except TokenError:
            pass
        else:
            if token.access_token:
                return True
        return False

    def _err(self, txt):
        logger.error(sanitize(txt))
        raise OIDCError(txt)

    def _do_code(self, response, authresp):
        """Perform code flow."""
        # get the access token
        try:
            args = {
                "code": authresp["code"],
                "redirect_uri": self.registration_response["redirect_uris"][0],
                "client_id": self.client_id,
                "client_secret": self.client_secret,
            }

            try:
                args["scope"] = response["scope"]
            except KeyError:
                pass

            atresp = self.do_access_token_request(
                state=authresp["state"],
                request_args=args,
                authn_method=self.registration_response["token_endpoint_auth_method"],
            )
            msg = "Access token response: {}"
            logger.info(msg.format(sanitize(atresp)))
        except Exception as err:
            logger.error("%s" % err)
            raise

        if isinstance(atresp, ErrorResponse):
            msg = "Error response: {}"
            self._err(msg.format(sanitize(atresp.to_dict())))

        _token = atresp["access_token"]

        _id_token = atresp.get("id_token")
        return _token, _id_token

    def callback(self, response, session, format="dict"):
        """
        Call when an AuthN response has been received from the OP.

        :param response: The URL returned by the OP
        :return:
        """
        try:
            authresp = self.parse_response(
                AuthorizationResponse, response, sformat=format, keyjar=self.keyjar
            )
        except ResponseError:
            msg = "Could not parse response: '{}'"
            logger.error(msg.format(sanitize(response)))
            raise OIDCError("Problem parsing response")

        logger.info("AuthorizationReponse: {}".format(sanitize(authresp)))
        if isinstance(authresp, ErrorResponse):
            if authresp["error"] == "login_required":
                return self.create_authn_request(session)
            else:
                raise OIDCError("Access denied")

        _state = authresp["state"]

        _id_token = authresp.get("id_token")
        if (
            _id_token is not None
            and _id_token["nonce"] != self.authz_req[_state]["nonce"]
        ):
            self._err("Received nonce not the same as expected.")

        if self.behaviour["response_type"] == "code":
            _token, new_id_token = self._do_code(response, authresp)
            if new_id_token is not None:
                _id_token = new_id_token
        else:
            _token = authresp["access_token"]

        if not self.oidc:
            return {"access_token": _token}

        if _id_token is None:
            self._err("Invalid response: no IdToken")

        if _id_token["iss"] != self.provider_info["issuer"]:
            self._err("Issuer mismatch")

        if _id_token["nonce"] != self.authz_req[_state]["nonce"]:
            self._err("Nonce mismatch")

        if not self.allow_sign_alg_none:
            if _id_token.jws_header["alg"] == "none":
                self._err('Do not allow "none" signature algorithm')

        user_id = "{}:{}".format(_id_token["iss"], _id_token["sub"])

        if self.get_userinfo:
            if self.userinfo_request_method:
                kwargs = {"method": self.userinfo_request_method}
            else:
                kwargs = {}

            if self.has_access_token(state=authresp["state"]):
                inforesp = self.do_user_info_request(state=authresp["state"], **kwargs)

                if isinstance(inforesp, ErrorResponse):
                    self._err("Invalid response %s." % inforesp["error"])

                userinfo = inforesp.to_dict()

                if _id_token["sub"] != userinfo["sub"]:
                    self._err("Invalid response: userid mismatch")

                logger.debug("UserInfo: %s" % sanitize(inforesp))

                try:
                    self.id_token[user_id] = _id_token
                except TypeError:
                    self.id_token = {user_id: _id_token}
            else:
                userinfo = {}
                for attr in OpenIDSchema.c_param:
                    try:
                        userinfo[attr] = _id_token[attr]
                    except KeyError:
                        pass

            return {
                "user_id": user_id,
                "userinfo": userinfo,
                "id_token": _id_token,
                "access_token": _token,
            }
        else:
            return {"user_id": user_id, "id_token": _id_token, "access_token": _token}


class OIDCClients(object):
    def __init__(self, config, base_url, seed="", jwks_info=None, verify_ssl=True):
        """
        Initialize the client.

        :param config: Imported configuration module
        :return:
        """
        self.client: Dict[str, Client] = {}
        self.client_cls = Client
        self.config = config
        self.seed = seed or rndstr(16)
        self.seed = self.seed.encode("utf8")
        self.path: Dict[str, str] = {}
        self.base_url = base_url
        self.jwks_info = jwks_info
        self.verify_ssl = verify_ssl

        for key, val in config.CLIENTS.items():
            if self.jwks_info:
                _val = val.copy()
                _val.update(self.jwks_info)
            else:
                _val = val

            if key == "":
                continue
            else:
                self.client[key] = self.create_client(**_val)

    def get_path(self, redirect_uris, issuer):
        for ruri in redirect_uris:
            p = urlsplit(ruri)
            self.path[p.path[1:]] = issuer

    def create_client(self, userid="", **kwargs):
        """
        Do an instantiation of a client instance.

        :param userid: An identifier of the user
        :param: Keyword arguments
            Keys are ["srv_discovery_url", "client_info", "client_registration",
            "provider_info"]
        :return: client instance
        """
        _key_set = set(list(kwargs.keys()))
        try:
            _verify_ssl = kwargs["verify_ssl"]
        except KeyError:
            _verify_ssl = self.verify_ssl
        else:
            _key_set.discard("verify_ssl")

        client = self.client_cls(
            client_authn_method=CLIENT_AUTHN_METHOD,
            behaviour=kwargs["behaviour"],
            verify_ssl=_verify_ssl,
        )

        try:
            client.userinfo_request_method = kwargs["userinfo_request_method"]
        except KeyError:
            pass
        else:
            _key_set.discard("userinfo_request_method")

        # The behaviour parameter is not significant for the election process
        _key_set.discard("behaviour")
        for param in ["allow"]:
            try:
                setattr(client, param, kwargs[param])
            except KeyError:
                pass
            else:
                _key_set.discard(param)

        if _key_set == {"client_info"}:  # Everything dynamic
            # There has to be a userid
            if not userid:
                raise MissingAttribute("Missing userid specification")

            # Find the service that provides information about the OP
            issuer = client.wf.discovery_query(userid)
            # Gather OP information
            client.provider_config(issuer)
            # register the client
            client.register(
                client.provider_info["registration_endpoint"], **kwargs["client_info"]
            )
            self.get_path(kwargs["client_info"]["redirect_uris"], issuer)
        elif _key_set == set(["client_info", "srv_discovery_url"]):
            # Ship the webfinger part
            # Gather OP information
            client.provider_config(kwargs["srv_discovery_url"])
            # register the client
            client.register(
                client.provider_info["registration_endpoint"], **kwargs["client_info"]
            )
            self.get_path(
                kwargs["client_info"]["redirect_uris"], kwargs["srv_discovery_url"]
            )
        elif _key_set == set(["provider_info", "client_info"]):
            client.handle_provider_config(
                ASConfigurationResponse(**kwargs["provider_info"]),
                kwargs["provider_info"]["issuer"],
            )
            client.register(
                client.provider_info["registration_endpoint"], **kwargs["client_info"]
            )

            self.get_path(
                kwargs["client_info"]["redirect_uris"],
                kwargs["provider_info"]["issuer"],
            )
        elif _key_set == set(["provider_info", "client_registration"]):
            client.handle_provider_config(
                ASConfigurationResponse(**kwargs["provider_info"]),
                kwargs["provider_info"]["issuer"],
            )
            client.store_registration_info(
                RegistrationResponse(**kwargs["client_registration"])
            )
            self.get_path(
                kwargs["client_info"]["redirect_uris"],
                kwargs["provider_info"]["issuer"],
            )
        elif _key_set == set(["srv_discovery_url", "client_registration"]):
            client.provider_config(kwargs["srv_discovery_url"])
            client.store_registration_info(
                RegistrationResponse(**kwargs["client_registration"])
            )
            self.get_path(
                kwargs["client_registration"]["redirect_uris"],
                kwargs["srv_discovery_url"],
            )
        else:
            raise Exception("Configuration error ?")

        return client

    def dynamic_client(self, userid="", issuer=""):
        client = self.client_cls(
            client_authn_method=CLIENT_AUTHN_METHOD,
            verify_ssl=self.verify_ssl,
            **self.jwks_info,
        )

        if userid:
            issuer = client.wf.discovery_query(userid)

        if not issuer:
            raise OIDCError("Missing issuer")

        logger.info("issuer: {}".format(issuer))
        if issuer in self.client:
            return self.client[issuer]
        else:
            # Gather OP information
            _pcr = client.provider_config(issuer)
            logger.info("Provider info: {}".format(sanitize(_pcr.to_dict())))
            # register the client
            _cinfo = self.config.CLIENTS[""]["client_info"]
            reg_args = copy.copy(_cinfo)
            h = hashlib.sha256(self.seed)
            h.update(issuer.encode("utf8"))  # issuer has to be bytes
            base_urls = _cinfo["redirect_uris"]

            reg_args["redirect_uris"] = [
                u.format(base=self.base_url, iss=h.hexdigest()) for u in base_urls
            ]
            try:
                reg_args["post_logout_redirect_uris"] = [
                    u.format(base=self.base_url, iss=h.hexdigest())
                    for u in reg_args["post_logout_redirect_uris"]
                ]
            except KeyError:
                pass

            self.get_path(reg_args["redirect_uris"], issuer)
            if client.jwks_uri:
                reg_args["jwks_uri"] = client.jwks_uri

            rr = client.register(_pcr["registration_endpoint"], **reg_args)
            msg = "Registration response: {}"
            logger.info(msg.format(sanitize(rr.to_dict())))

            try:
                client.behaviour.update(**self.config.CLIENTS[""]["behaviour"])
            except KeyError:
                pass

            self.client[issuer] = client
            return client

    def __getitem__(self, item):
        """
        Given a service or user identifier return a suitable client.

        :param item:
        :return:
        """
        try:
            return self.client[item]
        except KeyError:
            return self.dynamic_client(issuer=item)

    def keys(self):
        return list(self.client.keys())

    def return_paths(self):
        return self.path.keys()



import logging
import time
from typing import Any
from typing import Dict

from oic.utils.authn.user import ToOld
from oic.utils.http_util import CookieDealer
from oic.utils.sanitize import sanitize

logger = logging.getLogger(__name__)


class AuthzHandling(CookieDealer):
    """Class that allows an entity to manage authorization."""

    def __init__(self):
        self.permdb: Dict[str, Any] = {}

    def __call__(self, *args, **kwargs):
        return ""

    def permissions(self, cookie=None, **kwargs):
        if cookie is None:
            return None
        else:
            logger.debug("kwargs: %s" % sanitize(kwargs))

            val = self.getCookieValue(cookie, self.srv.cookie_name)
            if val is None:
                return None
            else:
                uid, _ts, typ = val

            if typ == "uam":  # short lived
                _now = int(time.time())
                if _now > (int(_ts) + int(self.cookie_ttl * 60)):
                    logger.debug("Authentication timed out")
                    raise ToOld(
                        "%d > (%d + %d)" % (_now, int(_ts), int(self.cookie_ttl * 60))
                    )
            else:
                if "max_age" in kwargs and kwargs["max_age"]:
                    _now = int(time.time())
                    if _now > (int(_ts) + int(kwargs["max_age"])):
                        logger.debug("Authentication too old")
                        raise ToOld(
                            "%d > (%d + %d)" % (_now, int(_ts), int(kwargs["max_age"]))
                        )

            return self.permdb[uid]


class UserInfoConsent(AuthzHandling):
    def __call__(self, user, userinfo, **kwargs):
        pass


class Implicit(AuthzHandling):
    def __init__(self, permission="implicit"):
        AuthzHandling.__init__(self)
        self.permission = permission

    def permissions(self, cookie=None, **kwargs):
        return self.permission



import json
import uuid
from typing import Dict

from oic.oauth2.message import OPTIONAL_LIST_OF_STRINGS
from oic.oauth2.message import SINGLE_REQUIRED_STRING
from oic.oauth2.message import Message
from oic.oic.message import SINGLE_REQUIRED_INT
from oic.utils.jwt import JWT
from oic.utils.sdb import Token
from oic.utils.time_util import utc_time_sans_frac

__author__ = "roland"


class TokenAssertion(Message):
    c_param = {
        "iss": SINGLE_REQUIRED_STRING,
        "azp": SINGLE_REQUIRED_STRING,
        "sub": SINGLE_REQUIRED_STRING,
        "kid": SINGLE_REQUIRED_STRING,
        "exp": SINGLE_REQUIRED_INT,
        "jti": SINGLE_REQUIRED_STRING,
        "aud": OPTIONAL_LIST_OF_STRINGS,  # Array of strings or string
    }


class JWTToken(Token, JWT):
    usage = "authorization_grant"

    def __init__(self, typ, keyjar, lt_pattern=None, extra_claims=None, **kwargs):
        self.type = typ
        Token.__init__(self, typ, **kwargs)
        kwargs.pop("token_storage", None)
        JWT.__init__(self, keyjar, msgtype=TokenAssertion, **kwargs)
        self.lt_pattern = lt_pattern or {}
        self.db: Dict[str, str] = {}
        self.session_info = {"": 600}
        self.exp_args = ["sinfo"]
        self.extra_claims = extra_claims or {}

    def __call__(self, sid, *args, **kwargs):
        """
        Return a token.

        :return:
        """
        try:
            _sinfo = kwargs["sinfo"]
        except KeyError:
            exp = self.do_exp(**kwargs)
            _tid = kwargs["target_id"]
        else:
            if "lifetime" in kwargs:
                _sinfo["lifetime"] = kwargs["lifetime"]
            exp = self.do_exp(**_sinfo)
            _tid = _sinfo["client_id"]
            if "scope" not in kwargs:
                _scope = None
                try:
                    _scope = _sinfo["scope"]
                except KeyError:
                    ar = json.loads(_sinfo["authzreq"])
                    try:
                        _scope = ar["scope"]
                    except KeyError:
                        pass
                if _scope:
                    kwargs["scope"] = " ".join(_scope)

            if self.usage == "authorization_grant":
                try:
                    kwargs["sub"] = _sinfo["sub"]
                except KeyError:
                    pass

            del kwargs["sinfo"]

        if "aud" in kwargs:
            if _tid not in kwargs["aud"]:
                kwargs["aud"].append(_tid)
        else:
            kwargs["aud"] = [_tid]

        if self.usage == "client_authentication":
            try:
                kwargs["sub"] = _tid
            except KeyError:
                pass
        else:
            if "azp" not in kwargs:
                kwargs["azp"] = _tid

        for param in ["lifetime", "grant_type", "response_type", "target_id"]:
            try:
                del kwargs[param]
            except KeyError:
                pass

        try:
            kwargs["kid"] = self.extra_claims["kid"]
        except KeyError:
            pass

        _jti = "{}-{}".format(self.type, uuid.uuid4().hex)
        _jwt = self.pack(jti=_jti, exp=exp, **kwargs)
        self.db[_jti] = sid
        return _jwt

    def do_exp(self, **kwargs):
        try:
            lifetime = kwargs["lifetime"]
        except KeyError:
            try:
                rt = " ".join(kwargs["response_type"])
            except KeyError:
                rt = " ".join(kwargs["grant_type"])

            try:
                lifetime = self.lt_pattern[rt]
            except KeyError:
                lifetime = self.lt_pattern[""]

        return utc_time_sans_frac() + lifetime

    def type_and_key(self, token):
        """
        Return type of Token (A=Access code, T=Token, R=Refresh token) and the session id.

        :param token: A token
        :return: tuple of token type and session id
        """
        msg = self.unpack(token)
        return self.type, self.db[msg["jti"]]

    def get_key(self, token):
        """
        Return session id.

        :param token: A token
        :return: The session id
        """
        msg = self.unpack(token)
        return self.db[msg["jti"]]

    def get_type(self, token):
        """
        Return token type.

        :param token: A token
        :return: Type of Token (A=Access code, T=Token, R=Refresh token)
        """
        self.unpack(token)
        return self.type

    def invalidate(self, token):
        info = self.unpack(token)
        try:
            del self.db[info["jti"]]
        except KeyError:
            return False

        return True

    def is_valid(self, info):
        if info["jti"] in self.db:
            if info["exp"] >= utc_time_sans_frac():
                return True

        return False

    def expires_at(self, token):
        info = self.unpack(token)
        return info["exp"]

    def valid(self, token):
        info = self.unpack(token)
        return self.is_valid(info)

    def get_info(self, token):
        return self.unpack(token)


class Authorization_Grant(JWTToken):
    usage = "authorization_grant"


class Client_Authentication(JWTToken):
    usage = "client_authentication"



import time
from typing import List
from typing import Optional

from oic.oauth2.message import AccessTokenResponse
from oic.oauth2.message import AuthorizationResponse
from oic.utils.time_util import utc_time_sans_frac

__author__ = "roland"


class Token(object):
    def __init__(self, resp=None):
        self.scope: List[str] = []
        self.token_expiration_time = 0
        self.access_token = None
        self.refresh_token = None
        self.token_type: Optional[str] = None
        self.replaced = False
        self.id_token = None

        if resp:
            for prop, val in resp.items():
                setattr(self, prop, val)

            try:
                _expires_in = resp["expires_in"]
            except KeyError:
                return

            if _expires_in:
                _tet = utc_time_sans_frac() + int(_expires_in)
            else:
                _tet = 0
            self.token_expiration_time = int(_tet)

    def is_valid(self):
        if self.token_expiration_time:
            if utc_time_sans_frac() > self.token_expiration_time:
                return False

        return True

    def __str__(self):
        return "%s" % self.__dict__

    def keys(self):
        return self.__dict__.keys()

    def __eq__(self, other):
        skeys = self.keys()
        okeys = other.keys()
        if set(skeys) != set(okeys):
            return False

        for key in skeys:
            if getattr(self, key) != getattr(other, key):
                return False

        return True


class Grant(object):
    _authz_resp = AuthorizationResponse
    _acc_resp = AccessTokenResponse
    _token_class = Token

    def __init__(self, exp_in=600, resp=None, seed=""):
        self.grant_expiration_time = 0
        self.exp_in = exp_in
        self.seed = seed
        self.tokens: List[Token] = []
        self.id_token = None
        self.code: Optional[str] = None
        if resp:
            self.add_code(resp)
            self.add_token(resp)

    @classmethod
    def from_code(cls, resp):
        instance = cls()
        instance.add_code(resp)
        return instance

    def add_code(self, resp):
        try:
            self.code = resp["code"]
            self.grant_expiration_time = utc_time_sans_frac() + self.exp_in
        except KeyError:
            pass

    def add_token(self, resp):
        """
        Add token to store.

        :param resp: An Authorization Response instance
        """
        if "access_token" in resp:
            tok = self._token_class(resp)
            self.tokens.append(tok)

    def is_valid(self):
        if utc_time_sans_frac() > self.grant_expiration_time:
            return False
        else:
            return True

    def __str__(self):
        return "%s" % self.__dict__

    def keys(self):
        return self.__dict__.keys()

    def update(self, resp):
        if "access_token" in resp or "id_token" in resp:
            tok = self._token_class(resp)
            if tok not in self.tokens:
                for otok in self.tokens:
                    otok.replaced = True
                self.tokens.append(tok)

        if "code" in resp:
            self.add_code(resp)

    def get_token(self, scope=""):
        token = None
        if scope:
            for token in self.tokens:
                if scope in token.scope and not token.replaced:
                    return token
        else:
            for token in self.tokens:
                if token.is_valid() and not token.replaced:
                    return token

        return token

    def delete_token(self, token):
        """Remove the specified token if it exists."""
        try:
            self.tokens.remove(token)
        except ValueError:
            pass

    def get_id_token(self):
        if self.id_token:
            if self.id_token["exp"] >= time.time():
                return self.id_token
        else:
            for tok in self.tokens:
                try:
                    _idt = tok.id_token
                except AttributeError:
                    pass
                else:
                    if _idt:
                        if _idt["exp"] >= time.time():
                            return _idt
        return None

    def join(self, grant):
        if not self.exp_in:
            self.exp_in = grant.exp_in
        if not self.grant_expiration_time:
            self.grant_expiration_time = grant.grant_expiration_time
        if not self.seed:
            self.seed = grant.seed
        for token in grant.tokens:
            if token not in self.tokens:
                for otok in self.tokens:
                    if token.scope == otok.scope:
                        otok.replaced = True
                self.tokens.append(token)



import hashlib
import string

# Since SystemRandom is not available on all systems
try:
    # Python 3.6+, designed for this usecase
    from secrets import choice
except ImportError:
    import random

    try:
        # Python 2.4+ if available on the platform
        _sysrand = random.SystemRandom()
        choice = _sysrand.choice
    except AttributeError:
        # Fallback, really bad
        import warnings

        choice = random.choice
        warnings.warn(
            "No good random number generator available on this platform. "
            "Security tokens will be weak and guessable.",
            RuntimeWarning,
        )

__author__ = "Roland Hedberg"
__version__ = "1.5.0"


OIDCONF_PATTERN = "%s/.well-known/openid-configuration"
CC_METHOD = {"S256": hashlib.sha256, "S384": hashlib.sha384, "S512": hashlib.sha512}


def rndstr(size=16):
    """
    Return a string of random ascii characters or digits.

    :param size: The length of the string
    :return: string
    """
    _basech = string.ascii_letters + string.digits
    return "".join([choice(_basech) for _ in range(size)])


BASECH = string.ascii_letters + string.digits + "-._~"


def unreserved(size=64):
    """
    Return a string of random ascii characters, digits and unreserved characters for use as RFC 7636 code verifiers.

    :param size: The length of the string
    :return: string
    """
    return "".join([choice(BASECH) for _ in range(size)])
